import pandas as pd
from pcnn.util import standardize, inverse_standardize, normalize, inverse_normalize

class DataSet:
    def __init__(self, data, interval, to_normalize, to_standardize) -> None:

        self.data = data
        self.interval = interval
        self.to_normalize = to_normalize
        self.to_standardize = to_standardize

        self.is_normalized = False
        self.is_standardized = False

        self.X = None
        self.Y = None
        self.min_ = None
        self.max_ = None
        self.mean = None
        self.std = None

    def standardize(self, data=None):
        """
        Function to standardize the dataset, i.e. put it to zero mean and 1 std
        The mean and std of each column (sensor) is kept in memory to reverse the
        operation and to be able to apply them to other datasets
        """

        inplace = False

        # If no data is provided, take the entire data of the class instance
        if data is None:
            data = self.data

            # First, check that no normalization or standardization was already performed
            assert not self.is_normalized, "The data is already normalized!"
            assert not self.is_standardized, "The data is already standardized!"

            # Set the standardized flag to true
            self.is_standardized = True
            inplace = True

        # Standardize the data, recall the mean and std
        data, mean, std = standardize(data)

        if inplace:
            self.data = data
            self.mean = mean
            self.std = std
        else:
            return data, mean, std

    def inverse_standardize(self, data=None, mean=None, std=None, inplace: bool = False):
        """
        Function to reverse the standardization to get the original scales back. If no data is provided,
        the entire dataset is scaled back.

        Args:
            data:       Data to inverse standardize
            mean:       Mean to use (takes the one of the dataset by default)
            std:        Std to use (takes the one of the dataset by default)
            inplace:    Flag to do the operation in place or not

        Returns:
            Standardized data if wanted
        """

        # First sanity check: the data is already standardized
        assert self.is_standardized, "The data is not standardized!"

        # If no data is provided, take the entire DataFrame of the DataSet
        if data is None:
            # First sanity check: the data is already normalized
            assert self.is_normalized, "The data is not normalized!"
            if inplace:
                self.is_standardized = False
                data_ = self.data
            else:
                data_ = self.data.copy()

        # Otherwise, copy the data to avoid issues
        else:
            if type(data) == pd.Series:
                data_ = pd.DataFrame(data=data, index=data.index, columns=[data.name])
            else:
                data_ = data.copy()

        if mean is None:
            mean = self.mean[data_.columns]
        if std is None:
            std = self.std[data_.columns]

        # Inverse the standardization with the wanted means and stds
        data = inverse_standardize(data=data_, mean=mean, std=std)

        # Return the scaled data if wanted
        if inplace:
            self.data = data
        else:
            return data

    def normalize(self, data=None):
        """
        Function to normalize the dataset, i.e. scale it by the min and max values
        The min and max of each column (sensor) is kept in memory to reverse the
        operation and to be able to apply them to other datasets

        There is an additional trick here, the data is actually scaled between
        0.1 and 0.9 instead of the classical 0-1 scaling to avoid saturation
        This is supposed to help the learning
        """
        inplace = False

        # If no data is provided, take the entire data
        if data is None:
            data = self.data

            # First, check that no normalization or standardization was already performed
            assert not self.is_standardized, "The data is already standardized!"
            assert not self.is_normalized, "The data is already normalized!"

            # Set the standardized flag to true
            self.is_normalized = True
            inplace = True

        # Normalize the data and recall mins and maxes
        data, min_, max_ = normalize(data)

        if inplace:
            self.data = data
            self.min_ = min_
            self.max_ = max_
        else:
            return data, min_, max_

    def inverse_normalize(self, data=None, min_=None, max_=None, inplace: bool = False):
        """
        Function to reverse the normalization to get the original scales back. If no data is provided,
        The entire data is scaled back.

        Args:
            data:       Data to inverse normalize
            min_:       Min to use (takes the one of the dataset by default)
            max_:       Max to use (takes the one of the dataset by default)
            inplace:    Flag to do the operation in place or not

        Returns:
            Normalized data if wanted
        """

        # If no data is provided, take the entire DataFrame of the DataSet
        if data is None:
            # First sanity check: the data is already normalized
            assert self.is_normalized, "The data is not normalized!"
            if inplace:
                self.is_normalized = False
                data_ = self.data
            else:
                data_ = self.data.copy()

        # Otherwise, copy the data to avoid issues
        else:
            if type(data) == pd.Series:
                data_ = pd.DataFrame(data=data, index=data.index, columns=[data.name])
            else:
                data_ = data.copy()

        if min_ is None:
            min_ = self.min_[data_.columns]
        if max_ is None:
            max_ = self.max_[data_.columns]

        # Inverse the normalization using the needed mins and maxes
        data = inverse_normalize(data=data_, min_=min_, max_=max_)

        # Return the scaled data if wanted
        if inplace:
            self.data = data
        else:
            return data

def prepare_data(data: pd.DataFrame, interval: int, model_kwargs: dict, Y_columns: list, X_columns: list = None, verbose: int = 2):
    """
    Pipeline of actions to take to prepare the data for some model.

    Args:
        data:                   DataFrame to use
        interval:               Sampling time (in minutes) of the data
        model_kwargs:           Model arguments (start and end date, interval, ...), see 'parameters.py'
        predict_differences:    Whether to predict differences in outputs
        Y_columns:              Name of the columns that are to be predicted
        X_columns:              Sensors (columns) of the input data, if None all columns are kept
        verbose:                Verbose of the function

    Returns:
        A preprocessed dataset ready to be put into a model
    """

    # Use the custom function to load and prepare the full dataset from the NEST data
    dataset = DataSet(data=data.copy(), interval=interval, to_normalize=model_kwargs['to_normalize'], 
                        to_standardize=model_kwargs['to_standardize'])

    if verbose > 0:
        print("\nPreparing the data...")

    # Reorder the columns to ensure they are in the right order
    dataset.X_columns = [x for x in dataset.data.columns if x in X_columns]
    dataset.Y_columns = Y_columns # [y for y in dataset.data.columns if y in Y_columns]

    # Only keep useful columns
    if X_columns is not None:
        to_keep = [x for x in dataset.data.columns if x in list(set(X_columns) | set(Y_columns))]
        dataset.data.drop(columns=[x for x in dataset.data.columns if x not in to_keep], inplace=True)

    # If standardization is wanted
    if dataset.to_standardize:
        if verbose > 0:
            print("Standardizing the data...")
        dataset.standardize()

    # Else, normalization is usually done
    elif dataset.to_normalize:
        if verbose > 0:
            print("Normalizing the data...")
        dataset.normalize()

    # Define inputs and labels
    if X_columns is not None:
        dataset.X = dataset.data[dataset.X_columns].iloc[:-1, :].copy().values
    else:
        dataset.X = dataset.data.iloc[:-1, :].copy().values
    dataset.Y = dataset.data[dataset.Y_columns].iloc[1:, :].copy().values

    assert len(dataset.X) == len(dataset.Y), "Something weird happened"

    # Print the result and return it
    if verbose > 0:
        print("Data ready!")

    return dataset



import torch
from torch import nn
import numpy as np


class PositiveLinear(nn.Module):
    """
    https://discuss.pytorch.org/t/positive-weights/19701/7
    """
    def __init__(self, in_features, out_features, require_bias=True):
        super(PositiveLinear, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.require_bias = require_bias
        self.log_weight = nn.Parameter(torch.Tensor(out_features, in_features))
        if self.require_bias:
            self.bias = nn.Parameter(torch.Tensor(1, out_features))
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.constant_(self.log_weight, 0.)

    def forward(self, input):
        if self.require_bias:
            return nn.functional.linear(input, self.log_weight.exp()) + self.bias
        else:
            return nn.functional.linear(input, self.log_weight.exp())


class PCNN(nn.Module):
    """
    S-PCNN model, from the paper
    `Towards Scalable Physically Consistent Neural Networks: an Application to Data-driven Multi-zone Thermal Building Models`
    L. Di Natale (1,2), B. Svetozarevic (1), P. Heer (1) and C.N. Jones (2)
    1 - Urban Energy Systems, Empa, Switzerland
    2 - Laboratoire d'Automatique 3, EPFL, Switzerland

    The PCNN has two parts to predict the temperature evolution in a zone of a building
      - A base module `D` that computes unknown effect using neural networks
      - A linear module, the energy accumulator `E`, which includes prior physical knowledge:
        - Energy inputs from the HVAC system increase (heating) or decrease (cooling) `E`
        - Heat losses are proportional to temperature gradients, both to the outside and neighboring zone

    The final temperature prediction is `T=D+E` at each step.
    `D` and `E` are then carried on to the next step and updated.
    """

    def __init__(
            self,
            device,
            inputs_D: list,
            learn_initial_hidden_states: bool,
            feed_input_through_nn: bool,
            input_nn_hidden_sizes: list,
            lstm_hidden_size: int,
            lstm_num_layers: int,
            layer_norm: bool,
            output_nn_hidden_sizes: list,
            case_column: int,
            temperature_column: int,
            out_column: int,
            neigh_column: int,
            power_column: int,
            zero_power: float,
            division_factor: list,
            normalization_variables: dict,
            parameter_scalings: dict,
    ):
        """
        Function to build the models.

        Args:
            device:                     CPU or GPU
            inputs_D:                   Input features used to predict the unforce dynamics (D)
            learn_initial_hidden_states:Whether to learn the initial hidden and cell states
            feed_input_through_nn:      Flag whether or not to preprocess the inputs before the LSTMs
            input_nn_hidden_sizes:      Hidden sizes of the input processing NNs
            lstm_hidden_size:           Hidden size of the LSTMs processing the input
            lstm_num_layers:            Number of layers for the LSTMs
            layer_norm:                 Flag whether to include a layer normalization layer after the LSTM
            output_nn_hidden_sizes:     Hidden sizes of the NNs
            case_column:                Index of the column corresponding to the case
            temperature_column:         Index of the column corresponding to the room temperature
            power_column:               Index of the column corresponding to the power consumption
            out_column:                 Index of the column corresponding to the outside temperature
            neigh_column:               Index of the column corresponding to the neighboring room temperature
            zero_power:                 Scaled value of zero heating/cooling power. This is needed since the
                                          model should predict no power consumption if the valves stay closed
            division_factor:            Factors to scale the base predictions and ensure not too big differences
                                          between timesteps
            normalization_variables:    Minimum and differences of various parameters in the data, so that we can
                                          unnormalize data when needed
            parameter_scalings:         Scalings of the parameters a, b, c, d in the network for a better learning
                                          Corresponds to a good initialization
        """

        # Start by initializing a PyTorch module
        super().__init__()

        # Recall the parameters for further use
        self.device = device
        self.inputs_D = inputs_D
        self.learn_initial_hidden_states = learn_initial_hidden_states
        self.feed_input_through_nn = feed_input_through_nn
        self.input_nn_hidden_sizes = input_nn_hidden_sizes
        self.lstm_hidden_size = lstm_hidden_size
        self.lstm_num_layers = lstm_num_layers
        self.layer_norm = layer_norm
        self.output_nn_hidden_sizes = output_nn_hidden_sizes
        self.case_column = case_column
        self.temperature_column = temperature_column
        self.power_column = power_column
        self.out_column = out_column
        self.neigh_column = neigh_column
        self.zero_power = torch.Tensor(zero_power).to(self.device)
        self.division_factor = torch.Tensor(division_factor).to(self.device)

        # Define latent variables
        self.last_D = None  ## D
        self.last_E = None  ## E

        # Recall normalization constants
        self.room_min = normalization_variables['Room'][0]
        self.room_diff = normalization_variables['Room'][1]
        self.neigh_min = normalization_variables['Neigh'][0]
        self.neigh_diff = normalization_variables['Neigh'][1]
        self.out_min = normalization_variables['Out'][0]
        self.out_diff = normalization_variables['Out'][1]

        # Specific scalings for the physical parameters `a`, `b`, `c`, `d`
        self.a_scaling = torch.Tensor(parameter_scalings['a']).to(self.device)
        self.b_scaling = torch.Tensor(parameter_scalings['b']).to(self.device)
        self.c_scaling = torch.Tensor(parameter_scalings['c']).to(self.device)
        self.d_scaling = torch.Tensor(parameter_scalings['d']).to(self.device)

        # Build the models
        self._build_model()

    def _build_model(self) -> None:
        """
        Function to build the model, i.e. a base temperature prediction `D` and then the added effect of
        heating/cooling and heat losses to the outside and neighboring zone in `E`.
            - `D` is composed of a LSTM, possibly followed by a NN and preceeded by another one,
                with a skip connection (ResNet)
            - `E` is a linear module with parameters `a`, `b`, `c`, `d`
        """

        ## Initialization of the parameters of `E`
        self.a = PositiveLinear(1, 1, require_bias=False)
        self.b = PositiveLinear(1, 1, require_bias=False)
        self.c = PositiveLinear(1, 1, require_bias=False)
        self.d = PositiveLinear(1, 1, require_bias=False)

        ## Initialization of `D`
        # Hidden and cell state initialization
        if self.learn_initial_hidden_states:
            self.initial_h = nn.Parameter(data=torch.zeros(self.lstm_num_layers, self.lstm_hidden_size))
            self.initial_c = nn.Parameter(data=torch.zeros(self.lstm_num_layers, self.lstm_hidden_size))

        # Process the input by a NN if wanted
        if self.feed_input_through_nn:
            size = [len(self.inputs_D)] + self.input_nn_hidden_sizes
            self.input_nn = nn.ModuleList([nn.Sequential(nn.Linear(size[i], size[i + 1]), nn.ReLU())
                                                for i in range(0, len(size) - 1)])

        # Create the LSTMs at the core of `D`, with normalization layers
        lstm_input_size = self.input_nn_hidden_sizes[-1] if self.feed_input_through_nn else len(self.inputs_D)
        self.lstm = nn.LSTM(input_size=lstm_input_size, hidden_size=self.lstm_hidden_size,
                                 num_layers=self.lstm_num_layers, batch_first=True)
        if self.layer_norm:
            self.norm = nn.LayerNorm(normalized_shape=self.lstm_hidden_size)

        # Create the NNs to process the output of the LSTMs for each modules 
        # ensure the last layer has size 1 since we only model one zone
        sizes = [self.lstm_hidden_size] + self.output_nn_hidden_sizes + [1]
        self.output_nn = nn.ModuleList([nn.Sequential(nn.Linear(sizes[i], sizes[i + 1]), nn.Tanh())
                                                for i in range(0, len(sizes) - 1)])

        # Xavier initialization of all the weights of NNs, parameters `a`, `b`, `c`, `d` are set to 1
        for name, param in self.named_parameters():
            if 'norm' not in name:
                if ('bias' in name) or ('log_weight' in name):
                    nn.init.constant_(param, 0.0)
                elif 'weight' in name:
                    nn.init.xavier_normal_(param)


    def forward(self, x_: torch.Tensor, states=None, warm_start: bool = False) -> dict:  # noqa: C901
        """
        Custom redefinition of the forward path. The only sepcifc thing is the warm start, where we feed
        true temperatures back to the model.
        Another trick: we actually predict the power along the temperature. This is a generalization, in the
        paper the power is known, so we just store one input as output. This is to be used when there is
        no direct control over the power input and it needs preprocessing through some 'g'.

        Args:
            x_:         Input
            states:     Original hidden and cell states if known
                          (for all LSTMs, i.e. the base, the heating and the cooling)
            warm_start: Whether we are warm starting the model
            mpc_mode:   Flag to pass to the MPC mode and return D and E separately

        Returns:
            The predicted temperature and power, and a tuple containing the hidden and cell states of
              the LSTMs
        """

        x = x_.clone()

        if len(x.shape) == 2:
            x = x.unsqueeze(1)

        # Take the given hidden and cell states of the LSTMs if they exist
        if states is not None:
            (h, c) = states

        # Otherwise, the sequence of data is new, so we need to initialize the hidden and cell states
        else:
            if self.learn_initial_hidden_states:
                h = torch.stack([self.initial_h.clone() for _ in range(x.shape[0])], dim=1).to(self.device)
                c = torch.stack([self.initial_c.clone() for _ in range(x.shape[0])], dim=1).to(self.device)
            else:
                h = torch.zeros((self.lstm_num_layers, x.shape[0], self.lstm_hidden_size)).to(self.device)
                c = torch.zeros((self.lstm_num_layers, x.shape[0], self.lstm_hidden_size)).to(self.device)

            # Define vectors to store 'D' and 'E', which will evolve through time
            self.last_D = torch.zeros((x.shape[0], 1)).to(self.device)  ## D
            self.last_E = torch.zeros((x.shape[0], 1)).to(self.device)  ## E

        # Use the last base predictions as input when not warm starting, otherwise we keep the true temperature
        if not warm_start:
            x[:, -1, self.temperature_column] = self.last_D
        else:
            x[:, -1, self.temperature_column] = x[:, -1, self.temperature_column].clone() - self.last_E

        ## Forward 'D'
        # Input embedding when wanted
        if self.feed_input_through_nn:
            D_embedding = torch.zeros(x.shape[0], x.shape[1], self.input_nn_hidden_sizes[-1]).to(self.device)
            for time_step in range(x.shape[1]):
                temp = x[:, time_step, self.inputs_D]
                for layer in self.input_nn:
                    temp = layer(temp)
                D_embedding[:, time_step, :] = temp
        else:
            D_embedding = x[:, :, self.inputs_D]

        # LSTM prediction for the base temperature
        lstm_output, (h, c) = self.lstm(D_embedding, (h, c))

        if self.layer_norm:
            lstm_output = self.norm(lstm_output)

        # Some manipulations are needed to feed the output through the neural network if wanted
        # Put the data is the form needed for the neural net
        temp = lstm_output[:, -1, :]
        # Go through the input layer of the NN
        for layer in self.output_nn:
            temp = layer(temp)
        D = temp / self.division_factor + x[:, -1, self.temperature_column]
        
        ## Heat losses computation in 'E'
        E = self.last_E.clone()

        # Loss to the outside is b*(T_k-T^out_k)
        if self.out_column is not None:
            E = E.clone() - self.b(
                ((x[:, -1, self.temperature_column] + self.last_E.clone() - 0.1) / 0.8
                  * self.room_diff + self.room_min)
                 - ((x[:, -1, [self.out_column]] - 0.1) / 0.8
                    * self.out_diff + self.out_min)) / self.b_scaling

        # Loss to the neighboring zone is c*(T_k-T^neigh_k)
        for i in range(len(self.neigh_column)):
            E = E.clone() - self.c(
                (((x[:, -1, self.temperature_column]+ self.last_E.clone() - 0.1) / 0.8
                  * self.room_diff + self.room_min)
                - ((x[:, -1, [self.neigh_column[i]]] - 0.1) / 0.8
                    * self.neigh_diff[i] + self.neigh_min[i]))) / self.c_scaling

        ## Heating/cooling effect of HVAC on 'E'

        # Find sequences in the batch where there actually is heating/cooling
        # Trick needed to put padded values back to zero power before detecting power intputs
        temp = torch.where(x[:, :, self.power_column] > 0.05, x[:, :, self.power_column], self.zero_power)
        mask = torch.where(torch.abs(temp - self.zero_power) > 1e-6, True, False).squeeze(1)

        if sum(mask) > 0:

            # Find heating and cooling sequences
            heating = x[:, 0, [self.case_column]] > 0.5
            cooling = x[:, 0, [self.case_column]] < 0.5

            # Substract the 'zero power' to get negative values for cooling power
            power = x[:, -1, self.power_column].clone() - self.zero_power

            if sum(heating) > 0:
                # Heating effect: add a*u to 'E'
                E[mask & heating] = E[mask & heating].clone() + self.a(power[mask & heating].unsqueeze(-1)).squeeze(-1) / self.a_scaling

            if sum(cooling) > 0:
                # Cooling effect: add d*u (where u<0 now, so we actually subtract energy) to 'E'
                E[mask & cooling] = E[mask & cooling].clone() + self.d(power[mask & cooling].unsqueeze(-1)).squeeze(-1) / self.d_scaling

        # Recall 'D' and 'E' for the next time step
        self.last_D = D.clone()
        self.last_E = E.clone()

        # Final computation of the result 'T=D+E'
        output = D + E
        # Trick needed since some sequences are padded
        output[torch.where(x[:, -1, 0] < 1e-6)[0], :] = 0.
        # Return the predictions and states of the model
        return output, (h, c)

    @property
    def E_parameters(self):
        return [float(x._parameters['log_weight']) for x in [self.a, self.b, self.c, self.d]]


class S_PCNN(nn.Module):
    """
    S-PCNN model, from the paper
    `Towards Scalable Physically Consistent Neural Networks: an Application to Data-driven Multi-zone Thermal Building Models`
    L. Di Natale (1,2), B. Svetozarevic (1), P. Heer (1) and C.N. Jones (2)
    1 - Urban Energy Systems, Empa, Switzerland
    2 - Laboratoire d'Automatique 3, EPFL, Switzerland

    The PCNN has two parts to predict the temperature evolution in a zone of a building
      - A base module `D` that computes unknown effect using neural networks
      - A linear module, the energy accumulator `E`, which includes prior physical knowledge:
        - Energy inputs from the HVAC system increase (heating) or decrease (cooling) `E`
        - Heat losses are proportional to temperature gradients, both to the outside and neighboring zone

    The final temperature prediction is `T=D+E` at each step.
    `D` and `E` are then carried on to the next step and updated.
    """

    def __init__(
            self,
            device,
            inputs_D: list,
            learn_initial_hidden_states: bool,
            feed_input_through_nn: bool,
            input_nn_hidden_sizes: list,
            lstm_hidden_size: int,
            lstm_num_layers: int,
            layer_norm: bool,
            output_nn_hidden_sizes: list,
            case_column: int,
            temperature_column: int,
            out_column: int,
            power_column: int,
            zero_power: float,
            division_factor: list,
            normalization_variables: dict,
            parameter_scalings: dict,
            topology: dict,
    ):
        """
        Function to build the models.

        Args:
            device:                     CPU or GPU
            inputs_D:                   Input features used to predict the unforce dynamics (D)
            learn_initial_hidden_states:Whether to learn the initial hidden and cell states
            feed_input_through_nn:      Flag whether or not to preprocess the inputs before the LSTMs
            input_nn_hidden_sizes:      Hidden sizes of the input processing NNs
            lstm_hidden_size:           Hidden size of the LSTMs processing the input
            lstm_num_layers:            Number of layers for the LSTMs
            layer_norm:                 Flag whether to include a layer normalization layer after the LSTM
            output_nn_hidden_sizes:     Hidden sizes of the NNs
            case_column:                Index of the column corresponding to the case
            temperature_column:         Index of the column corresponding to the room temperature
            power_column:               Index of the column corresponding to the power consumption
            out_column:                 Index of the column corresponding to the outside temperature
            zero_power:                 Scaled value of zero heating/cooling power. This is needed since the
                                          model should predict no power consumption if the valves stay closed
            division_factor:            Factors to scale the base predictions and ensure not too big differences
                                          between timesteps
            normalization_variables:    Minimum and differences of various parameters in the data, so that we can
                                          unnormalize data when needed
            parameter_scalings:         Scalings of the parameters a, b, c, d in the network for a better learning
                                          Corresponds to a good initialization
            topology:                   Topology of the building to model
        """

        # Start by initializing a PyTorch module
        super().__init__()

        # Recall the parameters for further use
        self.device = device
        self.inputs_D = inputs_D
        self.learn_initial_hidden_states = learn_initial_hidden_states
        self.feed_input_through_nn = feed_input_through_nn
        self.input_nn_hidden_sizes = input_nn_hidden_sizes
        self.lstm_hidden_size = lstm_hidden_size
        self.lstm_num_layers = lstm_num_layers
        self.layer_norm = layer_norm
        self.output_nn_hidden_sizes = output_nn_hidden_sizes
        self.case_column = case_column
        self.temperature_column = temperature_column
        self.power_column = power_column
        self.out_column = out_column
        self.zero_power = torch.Tensor(zero_power).to(self.device)
        self.division_factor = torch.Tensor(division_factor).to(self.device)
        self.topology = topology

        # Define latent variables
        self.last_D = None  ## D
        self.last_E = None  ## E

        # Recall normalization constants
        self.room_min = normalization_variables['Room'][0]
        self.room_diff = normalization_variables['Room'][1]
        self.out_min = normalization_variables['Out'][0]
        self.out_diff = normalization_variables['Out'][1]

        # Specific scalings for the physical parameters `a`, `b`, `c`, `d`
        self.a_scaling = torch.Tensor(parameter_scalings['a']).to(self.device)
        self.b_scaling = torch.Tensor(parameter_scalings['b']).to(self.device)
        self.c_scaling = torch.Tensor(parameter_scalings['c']).to(self.device)
        self.d_scaling = torch.Tensor(parameter_scalings['d']).to(self.device)

        # Build the models
        self._build_model()

    def _build_model(self) -> None:
        """
        Function to build the model, i.e. a base temperature prediction `D` and then the added effect of
        heating/cooling and heat losses to the outside and neighboring zone in `E`.
            - `D` is composed of a LSTM, possibly followed by a NN and preceeded by another one,
                with a skip connection (ResNet)
            - `E` is a linear module with parameters `a`, `b`, `c`, `d`
        """

        ## Initialization of the parameters of `E`
        self.a = nn.ModuleList(
            [PositiveLinear(1, 1, require_bias=False) for _ in range(len(self.topology['Rooms']))])
        self.b = nn.ModuleList(
            [PositiveLinear(1, 1, require_bias=False) for _ in range(len(self.topology['Outside']))])
        self.c = nn.ModuleList(
            [PositiveLinear(1, 1, require_bias=False) for _ in range(len(self.topology['Neighbors']))])
        self.d = nn.ModuleList(
            [PositiveLinear(1, 1, require_bias=False) for _ in range(len(self.topology['Rooms']))])

        ## Initialization of `D`
        # Hidden and cell state initialization
        if self.learn_initial_hidden_states:
            self.initial_h = nn.Parameter(data=torch.zeros(self.lstm_num_layers, self.lstm_hidden_size))
            self.initial_c = nn.Parameter(data=torch.zeros(self.lstm_num_layers, self.lstm_hidden_size))

        # Process the input by a NN if wanted
        if self.feed_input_through_nn:
            size = [len(self.inputs_D)] + self.input_nn_hidden_sizes
            self.input_nn = nn.ModuleList([nn.Sequential(nn.Linear(size[i], size[i + 1]), nn.ReLU())
                                                for i in range(0, len(size) - 1)])

        # Create the LSTMs at the core of `D`, with normalization layers
        lstm_input_size = self.input_nn_hidden_sizes[-1] if self.feed_input_through_nn else len(self.inputs_D)
        self.lstm = nn.LSTM(input_size=lstm_input_size, hidden_size=self.lstm_hidden_size,
                                 num_layers=self.lstm_num_layers, batch_first=True)
        if self.layer_norm:
            self.norm = nn.LayerNorm(normalized_shape=self.lstm_hidden_size)

        # Create the NNs to process the output of the LSTMs for each modules if wanted
        sizes = [self.lstm_hidden_size] + self.output_nn_hidden_sizes + [len(self.topology['Rooms'])]
        self.output_nn = nn.ModuleList([nn.Sequential(nn.Linear(sizes[i], sizes[i + 1]), nn.Tanh())
                                                for i in range(0, len(sizes) - 1)])

        # Xavier initialization of all the weights of NNs, parameters `a`, `b`, `c`, `d` are set to 1
        for name, param in self.named_parameters():
            if 'norm' not in name:
                if ('bias' in name) or ('log_weight' in name):
                    nn.init.constant_(param, 0.0)
                elif 'weight' in name:
                    nn.init.xavier_normal_(param)


    def forward(self, x_: torch.Tensor, states=None, warm_start: bool = False) -> dict:  # noqa: C901
        """
        Custom redefinition of the forward path. The only sepcifc thing is the warm start, where we feed
        true temperatures back to the model.
        Another trick: we actually predict the power along the temperature. This is a generalization, in the
        paper the power is known, so we just store one input as output. This is to be used when there is
        no direct control over the power input and it needs preprocessing through some 'g'.

        Args:
            x_:         Input
            states:     Original hidden and cell states if known
                          (for all LSTMs, i.e. the base, the heating and the cooling)
            warm_start: Whether we are warm starting the model
            mpc_mode:   Flag to pass to the MPC mode and return D and E separately

        Returns:
            The predicted temperature and power, and a tuple containing the hidden and cell states of
              the LSTMs
        """

        x = x_.clone()

        if len(x.shape) == 2:
            x = x.unsqueeze(1)

        # Take the given hidden and cell states of the LSTMs if they exist
        if states is not None:
            (h, c) = states

        # Otherwise, the sequence of data is new, so we need to initialize the hidden and cell states
        else:
            if self.learn_initial_hidden_states:
                h = torch.stack([self.initial_h.clone() for _ in range(x.shape[0])], dim=1).to(self.device)
                c = torch.stack([self.initial_c.clone() for _ in range(x.shape[0])], dim=1).to(self.device)
            else:
                h = torch.zeros((self.lstm_num_layers, x.shape[0], self.lstm_hidden_size)).to(self.device)
                c = torch.zeros((self.lstm_num_layers, x.shape[0], self.lstm_hidden_size)).to(self.device)

            # Define vectors to store 'D' and 'E', which will evolve through time
            self.last_D = torch.zeros((x.shape[0], len(self.topology['Rooms']))).to(self.device)  ## D
            self.last_E = torch.zeros((x.shape[0], len(self.topology['Rooms']))).to(self.device)  ## E

        # Use the last base predictions as input when not warm starting, otherwise we keep the true temperature
        if not warm_start:
            x[:, -1, self.temperature_column] = self.last_D
        else:
            x[:, -1, self.temperature_column] = x[:, -1, self.temperature_column].clone() - self.last_E

        ## Forward 'D'
        # Input embedding when wanted
        if self.feed_input_through_nn:
            D_embedding = torch.zeros(x.shape[0], x.shape[1], self.input_nn_hidden_sizes[-1]).to(self.device)
            for time_step in range(x.shape[1]):
                temp = x[:, time_step, self.inputs_D]
                for layer in self.input_nn:
                    temp = layer(temp)
                D_embedding[:, time_step, :] = temp
        else:
            D_embedding = x[:, :, self.inputs_D]

        # LSTM prediction for the base temperature
        lstm_output, (h, c) = self.lstm(D_embedding, (h, c))

        if self.layer_norm:
            lstm_output = self.norm(lstm_output)

        # Some manipulations are needed to feed the output through the neural network if wanted
        # Put the data is the form needed for the neural net
        temp = lstm_output[:, -1, :]
        # Go through the input layer of the NN
        for layer in self.output_nn:
            temp = layer(temp)
        D = temp / self.division_factor + x[:, -1, self.temperature_column]
        
        ## Heat losses computation in 'E'
        E = self.last_E.clone()  

        # Loss to the outside is b*(T_k-T^out_k)
        for i, room in enumerate(self.topology['Outside']):
            E[:, room] = E[:, room].clone() - self.b[i](
                (((x[:, -1, self.temperature_column[room]]
                   + self.last_E[:, room].clone() - 0.1) / 0.8
                  * self.room_diff[room] + self.room_min[room])
                 - ((x[:, -1, self.out_column] - 0.1) / 0.8
                    * self.out_diff + self.out_min)).
                    reshape(-1, 1)).squeeze() / self.b_scaling

        # Loss to the neighboring zone is c*(T_k-T^neigh_k)
        for i, (rooma, roomb) in enumerate(self.topology['Neighbors']):
            for room1, room2 in zip([rooma, roomb], [roomb, rooma]):
                E[:, room1] = E[:, room1].clone() - self.c[i](
                    (((x[:, -1, self.temperature_column[room1]]
                    + self.last_E[:, room1].clone() - 0.1) / 0.8
                    * self.room_diff[room1] + self.room_min[room1])
                    - ((x[:, -1, self.temperature_column[room2]]
                        + self.last_E[:, room2].clone() - 0.1) / 0.8
                        * self.room_diff[room2] + self.room_min[room2])).
                        reshape(-1, 1)).squeeze() / self.c_scaling

        ## Heating/cooling effect of HVAC on 'E'
        
        # Find sequences in the batch where there actually is heating/cooling
        # Trick needed to put padded values back to zero power before detecting power intputs
        temp = torch.where(x[:, :, self.power_column] > 0.05, x[:, :, self.power_column], self.zero_power)
        mask = torch.where(torch.abs(temp - self.zero_power) > 1e-6, True, False).squeeze(1).sum(axis=-1) > 0
        
        if sum(mask) > 0:

            # Find heating and cooling sequences
            heating = x[:, 0, self.case_column] > 0.5
            cooling = x[:, 0, self.case_column] < 0.5

            # Substract the 'zero power' to get negative values for cooling power
            power = x[:, -1, self.power_column].clone() - self.zero_power

            if sum(heating) > 0:
                # Heating effect: add a*u to 'E'
                for i in range(len(self.topology['Rooms'])):
                    E[mask & heating, i] = E[mask & heating, i].clone() + self.a[i](
                        power[mask & heating, i].unsqueeze(-1)).squeeze() / self.a_scaling[i]

            if sum(cooling) > 0:
                # Cooling effect: add d*u (where u<0 now, so we actually subtract energy) to 'E'
                for i in range(len(self.topology['Rooms'])):
                    E[mask & cooling, i] = E[mask & cooling, i].clone() + self.d[i](
                        power[mask & cooling, i].unsqueeze(-1)).squeeze() / self.d_scaling[i]

        # Recall 'D' and 'E' for the next time step
        self.last_D = D.clone()
        self.last_E = E.clone()

        # Final computation of the result 'T=D+E'
        output = D + E
        # Trick needed since some sequences are padded
        output[torch.where(x[:, -1, 0] < 1e-6)[0], :] = 0.

        # Return the predictions and states of the model
        return output, (h, c)

    @property
    def E_parameters(self):
        return [[float(y._parameters['log_weight']) for y in x] for x in [self.a, self.b, self.c, self.d]]



class M_PCNN(nn.Module):
    """
    S-PCNN model, from the paper
    `Towards Scalable Physically Consistent Neural Networks: an Application to Data-driven Multi-zone Thermal Building Models`
    L. Di Natale (1,2), B. Svetozarevic (1), P. Heer (1) and C.N. Jones (2)
    1 - Urban Energy Systems, Empa, Switzerland
    2 - Laboratoire d'Automatique 3, EPFL, Switzerland

    The PCNN has two parts to predict the temperature evolution in a zone of a building
      - A base module `D` that computes unknown effect using neural networks
      - A linear module, the energy accumulator `E`, which includes prior physical knowledge:
        - Energy inputs from the HVAC system increase (heating) or decrease (cooling) `E`
        - Heat losses are proportional to temperature gradients, both to the outside and neighboring zone

    The final temperature prediction is `T=D+E` at each step.
    `D` and `E` are then carried on to the next step and updated.
    """

    def __init__(
            self,
            device,
            inputs_D: list,
            learn_initial_hidden_states: bool,
            feed_input_through_nn: bool,
            input_nn_hidden_sizes: list,
            lstm_hidden_size: int,
            lstm_num_layers: int,
            layer_norm: bool,
            output_nn_hidden_sizes: list,
            case_column: int,
            temperature_column: int,
            out_column: int,
            power_column: int,
            zero_power: float,
            division_factor: list,
            normalization_variables: dict,
            parameter_scalings: dict,
            topology: dict,
    ):
        """
        Function to build the models.

        Args:
            device:                     CPU or GPU
            inputs_D:                   Input features used to predict the unforce dynamics (D)
            learn_initial_hidden_states:Whether to learn the initial hidden and cell states
            feed_input_through_nn:      Flag whether or not to preprocess the inputs before the LSTMs
            input_nn_hidden_sizes:      Hidden sizes of the input processing NNs
            lstm_hidden_size:           Hidden size of the LSTMs processing the input
            lstm_num_layers:            Number of layers for the LSTMs
            layer_norm:                 Flag whether to include a layer normalization layer after the LSTM
            output_nn_hidden_sizes:     Hidden sizes of the NNs
            case_column:                Index of the column corresponding to the case
            temperature_column:         Index of the column corresponding to the room temperature
            power_column:               Index of the column corresponding to the power consumption
            out_column:                 Index of the column corresponding to the outside temperature
            zero_power:                 Scaled value of zero heating/cooling power. This is needed since the
                                          model should predict no power consumption if the valves stay closed
            division_factor:            Factors to scale the base predictions and ensure not too big differences
                                          between timesteps
            normalization_variables:    Minimum and differences of various parameters in the data, so that we can
                                          unnormalize data when needed
            parameter_scalings:         Scalings of the parameters a, b, c, d in the network for a better learning
                                          Corresponds to a good initialization
            topology:                   Topology of the building to model
        """

        # Start by initializing a PyTorch module
        super().__init__()

        # Recall the parameters for further use
        self.device = device
        self.inputs_D = inputs_D
        self.learn_initial_hidden_states = learn_initial_hidden_states
        self.feed_input_through_nn = feed_input_through_nn
        self.input_nn_hidden_sizes = input_nn_hidden_sizes
        self.lstm_hidden_size = lstm_hidden_size
        self.lstm_num_layers = lstm_num_layers
        self.layer_norm = layer_norm
        self.output_nn_hidden_sizes = output_nn_hidden_sizes
        self.case_column = case_column
        self.temperature_column = temperature_column
        self.power_column = power_column
        self.out_column = out_column
        self.zero_power = torch.Tensor(zero_power).to(self.device)
        self.division_factor = torch.Tensor(division_factor).to(self.device)
        self.topology = topology

        # Define latent variables
        self.last_D = None  ## D
        self.last_E = None  ## E

        # Recall normalization constants
        self.room_min = normalization_variables['Room'][0]
        self.room_diff = normalization_variables['Room'][1]
        self.out_min = normalization_variables['Out'][0]
        self.out_diff = normalization_variables['Out'][1]

        # Specific scalings for the physical parameters `a`, `b`, `c`, `d`
        self.a_scaling = torch.Tensor(parameter_scalings['a']).to(self.device)
        self.b_scaling = torch.Tensor(parameter_scalings['b']).to(self.device)
        self.c_scaling = torch.Tensor(parameter_scalings['c']).to(self.device)
        self.d_scaling = torch.Tensor(parameter_scalings['d']).to(self.device)

        # Build the models
        self._build_model()

    def _build_model(self) -> None:
        """
        Function to build the model, i.e. a base temperature prediction `D` and then the added effect of
        heating/cooling and heat losses to the outside and neighboring zone in `E`.
            - `D` is composed of a LSTM, possibly followed by a NN and preceeded by another one,
                with a skip connection (ResNet)
            - `E` is a linear module with parameters `a`, `b`, `c`, `d`
        """

        ## Initialization of the parameters of `E`
        self.a = nn.ModuleList(
            [PositiveLinear(1, 1, require_bias=False) for _ in range(len(self.topology['Rooms']))])
        self.b = nn.ModuleList(
            [PositiveLinear(1, 1, require_bias=False) for _ in range(len(self.topology['Outside']))])
        self.c = nn.ModuleList(
            [PositiveLinear(1, 1, require_bias=False) for _ in range(len(self.topology['Neighbors']))])
        self.d = nn.ModuleList(
            [PositiveLinear(1, 1, require_bias=False) for _ in range(len(self.topology['Rooms']))])

        ## Initialization of `D`
        # Hidden and cell state initialization
        if self.learn_initial_hidden_states:
            self.initial_h = nn.ParameterList([nn.Parameter(data=torch.zeros(self.lstm_num_layers, self.lstm_hidden_size)) for _ in range(len(self.topology['Rooms']))])
            self.initial_c = nn.ParameterList([nn.Parameter(data=torch.zeros(self.lstm_num_layers, self.lstm_hidden_size)) for _ in range(len(self.topology['Rooms']))])

        # Process the input by a NN if wanted
        if self.feed_input_through_nn:
            sizes = [[len(input_D)] + self.input_nn_hidden_sizes for input_D in self.inputs_D]
            self.input_nn = nn.ModuleList([nn.ModuleList([nn.Sequential(nn.Linear(size[i], size[i + 1]), nn.ReLU())
                                                               for i in range(0, len(size) - 1)]) for size in sizes])

        # Create the LSTMs at the core of `D`, with normalization layers
        lstm_input_sizes = [self.input_nn_hidden_sizes[-1] if self.feed_input_through_nn else len(input_D) for
                           input_D in self.inputs_D] 
        self.lstm = nn.ModuleList([nn.LSTM(input_size=lstm_input_sizes[i], hidden_size=self.lstm_hidden_size,
                                                num_layers=self.lstm_num_layers, batch_first=True) for i in
                                        range(len(self.topology['Rooms']))])
        if self.layer_norm:
            self.norm = nn.ModuleList(
                [nn.LayerNorm(normalized_shape=self.lstm_hidden_size) for _ in range(len(self.topology['Rooms']))])

        # Create the NNs to process the output of the LSTMs for each modules if wanted
        sizes = [self.lstm_hidden_size] + self.output_nn_hidden_sizes + [1]
        self.output_nn = nn.ModuleList([nn.ModuleList([nn.Sequential(nn.Linear(sizes[i], sizes[i + 1]), nn.Tanh())
                                                            for i in range(0, len(sizes) - 1)]) for _ in
                                                range(len(self.topology['Rooms']))])

        # Xavier initialization of all the weights of NNs, parameters `a`, `b`, `c`, `d` are set to 1
        for name, param in self.named_parameters():
            if 'norm' not in name:
                if ('bias' in name) or ('log_weight' in name):
                    nn.init.constant_(param, 0.0)
                elif 'weight' in name:
                    nn.init.xavier_normal_(param)


    def forward(self, x_: torch.Tensor, states=None, warm_start: bool = False) -> dict:  # noqa: C901
        """
        Custom redefinition of the forward path. The only sepcifc thing is the warm start, where we feed
        true temperatures back to the model.
        Another trick: we actually predict the power along the temperature. This is a generalization, in the
        paper the power is known, so we just store one input as output. This is to be used when there is
        no direct control over the power input and it needs preprocessing through some 'g'.

        Args:
            x_:         Input
            states:     Original hidden and cell states if known
                          (for all LSTMs, i.e. the base, the heating and the cooling)
            warm_start: Whether we are warm starting the model
            mpc_mode:   Flag to pass to the MPC mode and return D and E separately

        Returns:
            The predicted temperature and power, and a tuple containing the hidden and cell states of
              the LSTMs
        """

        x = x_.clone()

        if len(x.shape) == 2:
            x = x.unsqueeze(1)

        # Take the given hidden and cell states of the LSTMs if they exist
        if states is not None:
            (h, c) = states

        # Otherwise, the sequence of data is new, so we need to initialize the hidden and cell states
        else:
            if self.learn_initial_hidden_states:
                h = [torch.stack([initial_h.clone() for _ in range(x.shape[0])], dim=1).to(self.device) for
                          initial_h in self.initial_h]
                c = [torch.stack([initial_c.clone() for _ in range(x.shape[0])], dim=1).to(self.device) for
                          initial_c in self.initial_c]
            else:
                h = [torch.zeros((self.lstm_num_layers, x.shape[0], self.lstm_hidden_size)).to(self.device) for _
                          in range(len(self.topology['Rooms']))]
                c = [torch.zeros((self.lstm_num_layers, x.shape[0], self.lstm_hidden_size)).to(self.device) for _
                          in range(len(self.topology['Rooms']))]

            # Define vectors to store 'D' and 'E', which will evolve through time
            self.last_D = torch.zeros((x.shape[0], len(self.topology['Rooms']))).to(self.device)  ## D
            self.last_E = torch.zeros((x.shape[0], len(self.topology['Rooms']))).to(self.device)  ## E

        # Use the last base predictions as input when not warm starting, otherwise we keep the true temperature
        if not warm_start:
            x[:, -1, self.temperature_column] = self.last_D
        else:
            x[:, -1, self.temperature_column] = x[:, -1, self.temperature_column].clone() - self.last_E

        ## Forward 'D'
        D = torch.zeros(x.shape[0], len(self.topology['Rooms'])).to(self.device) 

        # Input embedding when wanted
        if self.feed_input_through_nn:
            embeddings = []
            for i, input_nn in enumerate(self.input_nn):
                D_embedding = torch.zeros(x.shape[0], x.shape[1], self.input_nn_hidden_sizes[-1]).to(self.device)
                for time_step in range(x.shape[1]):
                    temp = x[:, time_step, self.inputs_D[i]]
                    for layer in input_nn:
                        temp = layer(temp)
                    D_embedding[:, time_step, :] = temp
                embeddings.append(D_embedding)
        else:
            D_embedding = [x[:, :, input_D] for input_D in self.inputs_D]

        # LSTM prediction for the base temperature
        lstm_outputs = []
        for i, lstm in enumerate(self.lstm):
            lstm_output, (h[i], c[i]) = lstm(embeddings[i], (h[i], c[i]))
            lstm_outputs.append(lstm_output)

        if self.layer_norm:
            for i in range(len(self.norm)):
                lstm_outputs[i] = self.norm[i](lstm_outputs[i])

        # Some manipulations are needed to feed the output through the neural network
        for j, output_nn in enumerate(self.output_nn):
            # Put the data is the form needed for the neural net
            temp = lstm_outputs[j][:, -1, :]
            # Go through the input layer of the NN
            for layer in output_nn:
                temp = layer(temp)
            D[:, j] = temp.squeeze() / self.division_factor + x[:, -1, self.temperature_column[j]]

        ## Heat losses computation in 'E'
        E = self.last_E.clone()

        # Loss to the outside is b*(T_k-T^out_k)
        for i, room in enumerate(self.topology['Outside']):
            E[:, room] = E[:, room].clone() - self.b[i](
                (((x[:, -1, self.temperature_column[room]]
                   + self.last_E[:, room].clone() - 0.1) / 0.8
                  * self.room_diff[room] + self.room_min[room])
                 - ((x[:, -1, self.out_column] - 0.1) / 0.8
                    * self.out_diff + self.out_min)).
                    reshape(-1, 1)).squeeze() / self.b_scaling

        # Loss to the neighboring zone is c*(T_k-T^neigh_k)
        for i, (rooma, roomb) in enumerate(self.topology['Neighbors']):
            for room1, room2 in zip([rooma, roomb], [roomb, rooma]):
                E[:, room1] = E[:, room1].clone() - self.c[i](
                    (((x[:, -1, self.temperature_column[room1]]
                    + self.last_E[:, room1].clone() - 0.1) / 0.8
                    * self.room_diff[room1] + self.room_min[room1])
                    - ((x[:, -1, self.temperature_column[room2]]
                        + self.last_E[:, room2].clone() - 0.1) / 0.8
                        * self.room_diff[room2] + self.room_min[room2])).
                        reshape(-1, 1)).squeeze() / self.c_scaling

        ## Heating/cooling effect of HVAC on 'E'
        # Find sequences in the batch where there actually is heating/cooling
        # Trick needed to put padded values back to zero power before detecting power intputs
        temp = torch.where(x[:, :, self.power_column] > 0.05, x[:, :, self.power_column], self.zero_power)
        mask = torch.where(torch.abs(temp - self.zero_power) > 1e-6, True, False).squeeze(1).sum(axis=-1) > 0

        if sum(mask) > 0:

            # Find heating and cooling sequences
            heating = x[:, 0, self.case_column] > 0.5
            cooling = x[:, 0, self.case_column] < 0.5

            # Substract the 'zero power' to get negative values for cooling power
            power = x[:, -1, self.power_column].clone() - self.zero_power

            if sum(heating) > 0:
                # Heating effect: add a*u to 'E'
                for i in range(len(self.topology['Rooms'])):
                    E[mask & heating, i] = E[mask & heating, i].clone() \
                                                  + self.a[i](power[mask & heating, i].unsqueeze(-1)).squeeze() \
                                                  / self.a_scaling[i]

            if sum(cooling) > 0:
                # Cooling effect: add d*u (where u<0 now, so we actually subtract energy) to 'E'
                for i in range(len(self.topology['Rooms'])):
                    E[mask & cooling, i] = E[mask & cooling, i].clone() \
                                                  + self.d[i](power[mask & cooling, i].unsqueeze(-1)).squeeze() \
                                                  / self.d_scaling[i]

        # Recall 'D' and 'E' for the next time step
        self.last_D = D.clone()
        self.last_E = E.clone()

        # Final computation of the result 'T=D+E'
        output = D + E
        # Trick needed since some sequences are padded
        output[torch.where(x[:, -1, 0] < 1e-6)[0], :] = 0.

        # Return the predictions and states of the model
        return output, (h, c)

    @property
    def E_parameters(self):
        return [[float(y._parameters['log_weight']) for y in x] for x in [self.a, self.b, self.c, self.d]]

class LSTM(nn.Module):
    """
    Classical LSTM model for comparison (with an encoder and decoder), from
    `Towards Scalable Physically Consistent Neural Networks: an Application to Data-driven Multi-zone Thermal Building Models`
    L. Di Natale (1,2), B. Svetozarevic (1), P. Heer (1) and C.N. Jones (2)
    1 - Urban Energy Systems, Empa, Switzerland
    2 - Laboratoire d'Automatique 3, EPFL, Switzerland
    """

    def __init__(
            self,
            device,
            rooms: list,
            inputs_D: list,
            learn_initial_hidden_states: bool,
            feed_input_through_nn: bool,
            input_nn_hidden_sizes: list,
            lstm_hidden_size: int,
            lstm_num_layers: int,
            layer_norm: bool,
            output_nn_hidden_sizes: list,
            temperature_column: list,
            power_column: list,
            division_factor: list,
    ):
        """
        Function to build the models.

        Args:
            device:                     CPU or GPU
            inputs_D:                   Input features used to predict the unforce dynamics (D)
            learn_initial_hidden_states:Whether to learn the initial hidden and cell states
            feed_input_through_nn:      Flag whether or not to preprocess the inputs before the LSTMs
            input_nn_hidden_sizes:      Hidden sizes of the input processing NNs
            lstm_hidden_size:           Hidden size of the LSTMs processing the input
            lstm_num_layers:            Number of layers for the LSTMs
            layer_norm:                 Flag whether to include a layer normalization layer after the LSTM
            output_nn_hidden_sizes:     Hidden sizes of the NNs
            temperature_column:         Index of the column corresponding to the room temperature
            power_column:               Index of the column corresponding to the power consumption
            division_factor:            Factors to scale the base predictions and ensure not too big differences
                                          between timesteps
        """

        # Start by initializing a PyTorch module
        super().__init__()

        # Recall the parameters for further use
        self.device = device
        self.rooms = rooms
        self.inputs_D = inputs_D
        self.learn_initial_hidden_states = learn_initial_hidden_states
        self.feed_input_through_nn = feed_input_through_nn
        self.input_nn_hidden_sizes = input_nn_hidden_sizes
        self.lstm_hidden_size = lstm_hidden_size
        self.lstm_num_layers = lstm_num_layers
        self.layer_norm = layer_norm
        self.output_nn_hidden_sizes = output_nn_hidden_sizes
        self.temperature_column = temperature_column
        self.power_columns = power_column
        self.division_factor = torch.Tensor(division_factor).to(self.device)

        # Define latent variables
        self.last = None  

        # Build the models
        self._build_model()

    def _build_model(self) -> None:
        """
        Function to build the model, i.e. a base temperature prediction `D` and then the added effect of
        heating/cooling and heat losses to the outside and neighboring zone in `E`.
            - `D` is composed of a LSTM, possibly followed by a NN and preceeded by another one,
                with a skip connection (ResNet)
            - `E` is a linear module with parameters `a`, `b`, `c`, `d`
        """

        # Hidden and cell state initialization
        if self.learn_initial_hidden_states:
            self.initial_h = nn.Parameter(data=torch.zeros(self.lstm_num_layers, self.lstm_hidden_size))
            self.initial_c = nn.Parameter(data=torch.zeros(self.lstm_num_layers, self.lstm_hidden_size))

        # Process the input by a NN if wanted
        if self.feed_input_through_nn:
            size = [len(self.inputs_D)] + self.input_nn_hidden_sizes
            self.input_nn = nn.ModuleList([nn.Sequential(nn.Linear(size[i], size[i + 1]), nn.ReLU())
                                                for i in range(0, len(size) - 1)])

        # Create the LSTMs at the core of `D`, with normalization layers
        lstm_input_size = self.input_nn_hidden_sizes[-1] if self.feed_input_through_nn else len(self.inputs_D)
        self.lstm = nn.LSTM(input_size=lstm_input_size, hidden_size=self.lstm_hidden_size,
                                 num_layers=self.lstm_num_layers, batch_first=True)
        if self.layer_norm:
            self.norm = nn.LayerNorm(normalized_shape=self.lstm_hidden_size)

        # Create the NNs to process the output of the LSTMs for each modules
        sizes = [self.lstm_hidden_size] + self.output_nn_hidden_sizes + [len(self.rooms)]
        self.output_nn = nn.ModuleList([nn.Sequential(nn.Linear(sizes[i], sizes[i + 1]), nn.Tanh())
                                                for i in range(0, len(sizes) - 1)])

        # Xavier initialization of all the weights of NNs, parameters `a`, `b`, `c`, `d` are set to 1
        for name, param in self.named_parameters():
            if 'norm' not in name:
                if 'bias' in name:
                    nn.init.constant_(param, 0.0)
                elif 'weight' in name:
                    nn.init.xavier_normal_(param)


    def forward(self, x_: torch.Tensor, states=None, warm_start: bool = False) -> dict:  # noqa: C901
        """
        Custom redefinition of the forward path. The only sepcifc thing is the warm start, where we feed
        true temperatures back to the model.
        Another trick: we actually predict the power along the temperature. This is a generalization, in the
        paper the power is known, so we just store one input as output. This is to be used when there is
        no direct control over the power input and it needs preprocessing through some 'g'.

        Args:
            x_:         Input
            states:     Original hidden and cell states if known
                          (for all LSTMs, i.e. the base, the heating and the cooling)
            warm_start: Whether we are warm starting the model
            mpc_mode:   Flag to pass to the MPC mode and return D and E separately

        Returns:
            The predicted temperature and power, and a tuple containing the hidden and cell states of
              the LSTMs
        """

        x = x_.clone()

        if len(x.shape) == 2:
            x = x.unsqueeze(1)

        # Take the given hidden and cell states of the LSTMs if they exist
        if states is not None:
            (h, c) = states

        # Otherwise, the sequence of data is new, so we need to initialize the hidden and cell states
        else:
            if self.learn_initial_hidden_states:
                h = torch.stack([self.initial_h.clone() for _ in range(x.shape[0])], dim=1).to(self.device)
                c = torch.stack([self.initial_c.clone() for _ in range(x.shape[0])], dim=1).to(self.device)
            else:
                h = torch.zeros((self.lstm_num_layers, x.shape[0], self.lstm_hidden_size)).to(self.device)
                c = torch.zeros((self.lstm_num_layers, x.shape[0], self.lstm_hidden_size)).to(self.device)

            # Define vectors to store 'D' and 'E', which will evolve through time
            self.last = torch.zeros((x.shape[0], len(self.temperature_column))).to(self.device)  ## D

        # Use the last base predictions as input when not warm starting, otherwise we keep the true temperature
        if not warm_start:
            x[:, -1, self.temperature_column] = self.last

        ## Forward 'D'
        # Input embedding when wanted
        if self.feed_input_through_nn:
            D_embedding = torch.zeros(x.shape[0], x.shape[1], self.input_nn_hidden_sizes[-1]).to(self.device)
            for time_step in range(x.shape[1]):
                temp = x[:, time_step, self.inputs_D]
                for layer in self.input_nn:
                    temp = layer(temp)
                D_embedding[:, time_step, :] = temp
        else:
            D_embedding = x[:, :, self.inputs_D]

        # LSTM prediction for the base temperature
        lstm_output, (h, c) = self.lstm(D_embedding, (h, c))

        if self.layer_norm:
            lstm_output = self.norm(lstm_output)

        # Some manipulations are needed to feed the output through the neural network if wanted
        # Put the data is the form needed for the neural net
        temp = lstm_output[:, -1, :]
        # Go through the input layer of the NN
        for layer in self.output_nn:
            temp = layer(temp)
        output = temp.squeeze() / self.division_factor + x[:, -1, self.temperature_column]

        # Recall 'D' and 'E' for the next time step
        self.last = output.clone()

        # Trick needed since some sequences are padded
        output[torch.where(x[:, -1, 0] < 1e-6)[0], :] = 0.

        # Return the predictions and states of the model
        return output, (h, c)



import os
import pandas as pd
import numpy as np
from typing import Union

from pcnn.parameters import DATA_SAVE_PATH

def load_data(save_name: str, save_path: str = DATA_SAVE_PATH) -> pd.DataFrame:
    """
    Function to load a dataframe if it exists

    Args:
        save_name:  Name of the file to load
        save_path:  Where to load it
    """

    # Build the full path and check its existence
    full_path = os.path.join(save_path, save_name + ".csv")
    assert os.path.exists(full_path), f"The file {full_path} doesn't exist."

    # Load the data and change the index to timestamps
    data = pd.read_csv(full_path, index_col=[0])
    data.index = pd.to_datetime(data.index)

    return data

def standardize(data: pd.DataFrame):
    """
    function to standardize the columns of a DataFrame.

    Args:
        data: The DataFrame to standardize

    Returns:
        the standardized data, the means and the stds
    """

    # Define and keep the means and stds in memory
    mean = data.mean()
    std = data.std()

    # Substract the mean
    data = data.subtract(mean)

    # Little trick to handle constant data (zero std --> cannot divide)
    # Only consider non zero variance columns
    non_zero_std = np.where(data.std().values > 1e-10)[0]
    # If there is a zero variance column, it is actually useless since it is constant
    if len(non_zero_std) < len(data.columns):
        print(f"Warning, 0 std for columns {data.columns[np.where(data.std().values < 1e-10)[0]].values}, really useful?")

    # Divide by the std where possible
    data.iloc[:, non_zero_std] = data.iloc[:, non_zero_std].divide(std[non_zero_std])

    return data, mean, std


def inverse_standardize(data: pd.DataFrame, mean: pd.Series, std: pd.Series):
    """
    function to inverse the standardization of the columns of a DataFrame.

    Args:
        data:   The DataFrame to inverse standardize
        mean:   The mean values of the columns
        std:    The standard deviations of the columns

    Returns:
        The original data
    """

    # If the given data is a number
    if isinstance(data, float):
        # Can get back to the original scale back
        data = mean + (data * std)

    else:
        # Get the places where the variance is not zero and multiply back
        # (the other columns were ignored)
        non_zero_std = np.where(std.values > 1e-10)[0]

        # In the case of an array
        if isinstance(data, np.ndarray):
            data[:, non_zero_std] = data[:, non_zero_std] * std[non_zero_std].values.reshape(1, -1)
            data += mean

        # If the given data is an DataFrame
        elif isinstance(data, pd.DataFrame):
            data.iloc[:, non_zero_std] = data.iloc[:, non_zero_std].multiply(std[non_zero_std])
            # Add the mean back
            data = data.add(mean)

        else:
            raise ValueError(f"Unexpected data type {type(data)}")

    return data


def normalize(data: pd.DataFrame):
    """
    Function to normalize the columns of a DataFrame to 0.1-0.9

    Args:
        data: The DataFrame to standardize

    Returns:
        the normalized data, the mins and the maxs
    """

    data = data.astype(float)

    # Define and save the min and max of each column
    max_ = data.max()
    min_ = data.min()

    # Little trick to handle constant data (zero std --> cannot divide)
    # Only consider non zero variance columns
    non_zero_div = np.where(max_ - min_ > 1e-10)[0]
    # If there is a zero variance column, it is actually useless since it is constant
    if len(non_zero_div) < len(data.columns):
        print(f"Warning, columns {data.columns[np.where(max_ - min_ < 1e-10)[0]].values} are constant, really useful?")

    # Scale the data between 0.1 and 0.9
    data.iloc[:, non_zero_div] = 0.8 * (data.iloc[:, non_zero_div] - min_.iloc[non_zero_div]) / (max_.iloc[non_zero_div] - min_.iloc[non_zero_div]) + 0.1

    data.iloc[:, np.where(max_ - min_ <= 1e-10)[0]] = 0.5

    return data, min_, max_


def inverse_normalize(data: Union[np.ndarray, pd.DataFrame, float], min_: Union[pd.Series, float], max_: Union[pd.Series, float]):
    """
    Function to inverse the normalization of the columns of a DataFrame to 0.1-0.9

    Args:
        data:   The DataFrame to inverse standardize
        min_:   The min values of the columns
        max_:   The max values of the columns

    Returns:
        The original data
    """

    # If the given data is a number
    if isinstance(data, float):
        # Can get back to the original scale back
        data = min_ + (data - 0.1) * (max_ - min_) / 0.8

    else:
        # Get the places where the variance is not zero and multiply back
        # (the other columns were ignored)
        non_zero_div = np.where(max_ - min_ > 1e-10)[0]

        # If the given data is an array
        if isinstance(data, np.ndarray):
            # Can get back to the original scale back
            data[:, non_zero_div] = (data[:, non_zero_div] - 0.1) * (max_[non_zero_div] - min_[non_zero_div]).values.reshape(1, -1) / 0.8
            data[:, non_zero_div] = data[:, non_zero_div] + min_[non_zero_div].values.reshape(1, -1)

        # If the given data is a DataFrame
        elif isinstance(data, pd.DataFrame):
            # Can get back to the original scale back
            data.iloc[:, non_zero_div] = (data.iloc[:, non_zero_div] - 0.1).multiply(max_.iloc[non_zero_div] - min_.iloc[non_zero_div]) / 0.8
            data.iloc[:, non_zero_div] = data.iloc[:, non_zero_div].add(min_.iloc[non_zero_div])

        else:
            raise ValueError(f"Unexpected data type {type(data)}")

    return data

def model_save_name_factory(rooms, module, model_kwargs):
    """
    Function to create helpful and somewhat unique names to easily save and load the wanted models
    This uses the starting and ending date of the data used to fit the model, as well as another
    part in "model_name" that is specific to each model, representing the model type and possibly
    some hyperparameters' choices

    Args:
        rooms:          List of rooms modeled
        module:         Module used in the model
        model_kwargs:   Parameters of the model, see 'parameters.py'

    Returns:
        A full name to save the model as
    """

    name = ''
    for room in rooms:
        name += f"{room}_"
    name = os.path.join(model_kwargs["save_path"], f"{model_kwargs['name']}_{model_kwargs['seed']}_"
                                                   f"{module}_{name}_")

    name += f"_{model_kwargs['warm_start_length']}_{model_kwargs['maximum_sequence_length']}"

    if (model_kwargs["heating"]) & (not model_kwargs["cooling"]):
        name += "_heating"
    elif (model_kwargs["cooling"]) & (not model_kwargs["heating"]):
        name += "_cooling"

    return name


def format_elapsed_time(tic, toc):
    """
    Small function to print the time elapsed between tic and toc in a nice manner
    """

    diff = toc - tic
    hours = int(diff // 3600)
    minutes = int((diff - 3600 * hours) // 60)
    seconds = str(int(diff - 3600 * hours - 60 * minutes))

    # Put everything in strings
    hours = str(hours)
    minutes = str(minutes)

    # Add a zero for one digit numbers for consistency
    if len(hours) == 1:
        hours = '0' + hours
    if len(minutes) == 1:
        minutes = '0' + minutes
    if len(seconds) == 1:
        seconds = '0' + seconds

    # Final nice looking print
    return f"{hours}:{minutes}:{seconds}"

def load_data(save_name: str, save_path: str = DATA_SAVE_PATH) -> pd.DataFrame:
    """
    Function to load a dataframe if it exists

    Args:
        save_name:  Name of the file to load
        save_path:  Where to load it
    """

    # Build the full path and check its existence
    full_path = os.path.join(save_path, save_name + ".csv")
    assert os.path.exists(full_path), f"The file {full_path} doesn't exist."

    # Load the data and change the index to timestamps
    data = pd.read_csv(full_path, index_col=[0])
    data.index = pd.to_datetime(data.index)

    return data


"""
File containing the base class of models, with general functions
"""

import os
import pandas as pd
import math
import time
from typing import Union

import random
import numpy as np
import matplotlib.pyplot as plt

import torch
from torch.nn.utils.rnn import pad_sequence
from torch import optim
import torch.nn.functional as F

from pcnn.module import PCNN, S_PCNN, M_PCNN, LSTM
from pcnn.data import prepare_data
from pcnn.util import model_save_name_factory, format_elapsed_time, inverse_normalize, inverse_standardize


class Model:
    """
    Class of models using PyTorch
    """

    def __init__(self, data: pd.DataFrame, interval: int, model_kwargs: dict, inputs_D: list, 
                module, rooms, case_column, out_column, neigh_column, temperature_column, power_column,
                Y_columns: list, X_columns: list = None, topology: dict = None, load_last: bool = False,
                load: bool = True, device: str = None):
        """
        Initialize a model.

        Args:
            model_kwargs:   Parameters of the models, see 'parameters.py'
            Y_columns:      Name of the columns that are to be predicted
            X_columns:      Sensors (columns) of the input data
        """


        assert module in ['PCNN', 'S_PCNN', 'M_PCNN', 'LSTM'],\
            f"The provided model type {module} does not exist, please chose among `'PCNN', 'S_PCNN', 'M_PCNN', 'LSTM'`."

        # Define the main attributes
        self.name = model_kwargs["name"]
        self.model_kwargs = model_kwargs
        self.rooms = rooms if isinstance(rooms, list) else [rooms]

        # Create the name associated to the model
        self.save_name = model_save_name_factory(rooms=self.rooms, module=module, model_kwargs=model_kwargs)

        if not os.path.isdir(self.save_name):
            os.mkdir(self.save_name)

        # Fix the seeds for reproduction
        self._fix_seeds(seed=model_kwargs["seed"])

        self.case_column = case_column
        self.out_column = out_column
        if neigh_column is not None:
            self.neigh_column = neigh_column if isinstance(neigh_column, list) else [neigh_column]
        else:
            self.neigh_column = neigh_column
        self.temperature_column = temperature_column if isinstance(temperature_column, list) else [temperature_column]
        self.power_column = power_column if isinstance(power_column, list) else [power_column]
        self.inputs_D = inputs_D
        self.topology = topology
        self.module = module

        self.unit = model_kwargs['unit']
        self.batch_size = model_kwargs["batch_size"]
        self.shuffle = model_kwargs["shuffle"]
        self.n_epochs = model_kwargs["n_epochs"]
        self.verbose = model_kwargs["verbose"]
        self.learning_rate = model_kwargs["learning_rate"]
        self.decrease_learning_rate = model_kwargs["decrease_learning_rate"]
        self.heating = model_kwargs["heating"]
        self.cooling = model_kwargs["cooling"]
        self.warm_start_length = model_kwargs["warm_start_length"]
        self.minimum_sequence_length = model_kwargs["minimum_sequence_length"]
        self.maximum_sequence_length = model_kwargs["maximum_sequence_length"]
        self.overlapping_distance = model_kwargs["overlapping_distance"]
        self.validation_percentage = model_kwargs["validation_percentage"]
        self.test_percentage = model_kwargs["test_percentage"]
        self.feed_input_through_nn = model_kwargs["feed_input_through_nn"]
        self.input_nn_hidden_sizes = model_kwargs["input_nn_hidden_sizes"]
        self.lstm_hidden_size = model_kwargs["lstm_hidden_size"]
        self.lstm_num_layers = model_kwargs["lstm_num_layers"]
        self.layer_norm = model_kwargs["layer_norm"]
        self.output_nn_hidden_sizes = model_kwargs["output_nn_hidden_sizes"]
        self.learn_initial_hidden_states = model_kwargs["learn_initial_hidden_states"]
        self.division_factor = model_kwargs['division_factor']
        self.model_kwargs = model_kwargs
   
        # Prepare the data
        self.dataset = prepare_data(data=data, interval=interval, model_kwargs=model_kwargs, 
                                    Y_columns=Y_columns, X_columns=X_columns, verbose=self.verbose)

        self.model = None
        self.optimizer = None
        self.loss = None
        self.train_losses = []
        self.validation_losses = []
        self._validation_losses = []
        self.test_losses = []
        self.a = []
        self.b = []
        self.c = []
        self.d = []
        self.times = []
        self.heating_sequences, self.cooling_sequences = None, None
        self.train_sequences = None
        self.validation_sequences = None
        self.test_sequences = None

        # Sanity check
        if self.neigh_column is None:
            print('\nSanity check of the columns:\n', [(w, [self.dataset.X_columns[i] for i in x]) 
                    for w, x in zip(['Case', 'Room temp', 'Room power', 'Out temp'],
                                    [[self.case_column], self.temperature_column, self.power_column,
                                     [self.out_column]])])
        else:
            print('\nSanity check of the columns:\n', [(w, [self.dataset.X_columns[i] for i in x]) 
                    for w, x in zip(['Case', 'Room temp', 'Room power', 'Out temp', 'Neigh temp'],
                                    [[self.case_column], self.temperature_column, self.power_column,
                                     [self.out_column], self.neigh_column])])

        print("Inputs used in D:\n", np.array(self.dataset.X_columns)[inputs_D])

        # To use the GPU when available
        if device is None:
            if torch.cuda.is_available():
                self.device = torch.device("cuda:0")
                print("\nGPU acceleration on!")
            elif torch.backends.mps.is_available():
                self.device = 'mps'
                print("Using mps.")
            else:
                self.device = "cpu"
                self.save_path = model_kwargs["save_path"]
        else:
            self.device = device

        # Compute the scaled zero power points and the division factors to use in ResNet-like
        # modules
        self.zero_power = self.compute_zero_power()
        self.normalization_variables = self.get_normalization_variables()
        self.parameter_scalings = self.create_scalings()

        # Prepare the torch module
        if self.module == "PCNN":
            self.model = PCNN(
                device=self.device,
                inputs_D=self.inputs_D,
                learn_initial_hidden_states=self.learn_initial_hidden_states,
                feed_input_through_nn=self.feed_input_through_nn,
                input_nn_hidden_sizes=self.input_nn_hidden_sizes,
                lstm_hidden_size=self.lstm_hidden_size,
                lstm_num_layers=self.lstm_num_layers,
                layer_norm=self.layer_norm,
                output_nn_hidden_sizes=self.output_nn_hidden_sizes,
                case_column=self.case_column,
                temperature_column=self.temperature_column,
                power_column=self.power_column,
                out_column=self.out_column,
                neigh_column=self.neigh_column,
                zero_power=self.zero_power,
                division_factor=self.division_factor,
                normalization_variables=self.normalization_variables,
                parameter_scalings=self.parameter_scalings,
            )
        
        elif self.module == "S_PCNN":
            self.model = S_PCNN(
                device=self.device,
                inputs_D=self.inputs_D,
                learn_initial_hidden_states=self.learn_initial_hidden_states,
                feed_input_through_nn=self.feed_input_through_nn,
                input_nn_hidden_sizes=self.input_nn_hidden_sizes,
                lstm_hidden_size=self.lstm_hidden_size,
                lstm_num_layers=self.lstm_num_layers,
                layer_norm=self.layer_norm,
                output_nn_hidden_sizes=self.output_nn_hidden_sizes,
                case_column=self.case_column,
                temperature_column=self.temperature_column,
                power_column=self.power_column,
                out_column=self.out_column,
                zero_power=self.zero_power,
                division_factor=self.division_factor,
                normalization_variables=self.normalization_variables,
                parameter_scalings=self.parameter_scalings,
                topology=self.topology,
            )

        elif self.module == "M_PCNN":
            self.model = M_PCNN(
                device=self.device,
                inputs_D=self.inputs_D,
                learn_initial_hidden_states=self.learn_initial_hidden_states,
                feed_input_through_nn=self.feed_input_through_nn,
                input_nn_hidden_sizes=self.input_nn_hidden_sizes,
                lstm_hidden_size=self.lstm_hidden_size,
                lstm_num_layers=self.lstm_num_layers,
                layer_norm=self.layer_norm,
                output_nn_hidden_sizes=self.output_nn_hidden_sizes,
                case_column=self.case_column,
                temperature_column=self.temperature_column,
                power_column=self.power_column,
                out_column=self.out_column,
                zero_power=self.zero_power,
                division_factor=self.division_factor,
                normalization_variables=self.normalization_variables,
                parameter_scalings=self.parameter_scalings,
                topology=self.topology,
            )

        elif self.module == "LSTM":
            self.model = LSTM(
                device=self.device,
                rooms=self.rooms,
                inputs_D=self.inputs_D,
                learn_initial_hidden_states=self.learn_initial_hidden_states,
                feed_input_through_nn=self.feed_input_through_nn,
                input_nn_hidden_sizes=self.input_nn_hidden_sizes,
                lstm_hidden_size=self.lstm_hidden_size,
                lstm_num_layers=self.lstm_num_layers,
                layer_norm=self.layer_norm,
                output_nn_hidden_sizes=self.output_nn_hidden_sizes,
                temperature_column=self.temperature_column,
                power_column=self.power_column,
                division_factor=self.division_factor
            )

        self.optimizer = optim.Adam(self.model.parameters(), lr=model_kwargs["learning_rate"])
        self.loss = F.mse_loss

        # Load the model if it exists
        if load:
            self.load(load_last=load_last)

        # if the model doesn't exist, the sequences were not loaded
        if self.train_sequences is None:
            self.heating_sequences, self.cooling_sequences = self.get_sequences()
            self.train_test_validation_separation(validation_percentage=self.validation_percentage,
                                                  test_percentage=self.test_percentage)

        self.model = self.model.to(self.device)

    @property
    def X(self):
        return self.dataset.X

    @property
    def Y(self):
        return self.dataset.Y

    @property
    def columns(self):
        return self.dataset.data.columns

    @property
    def differences_Y(self):
        return self.dataset.differences_Y

    def _fix_seeds(self, seed: int = None):
        """
        Function fixing the seeds for reproducibility.

        Args:
            seed:   Seed to fix everything
        """
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)

    def _create_sequences(self, X: pd.DataFrame = None, Y: pd.DataFrame = None, inplace: bool = False):
        """
        Function to create tuple designing the beginning and end of sequences of data we can predict.
        This is needed because PyTorch models don't work with NaN values, so we need to only path
        sequences of data that don't contain any.

        Args:
            X:          input data
            Y:          output data, i.e. labels
            inplace:    Flag whether to do it in place or not

        Returns:
            The created sequences if not inplace.
        """

        # Take the data of the current model if nothing is given
        if X is None:
            X = self.X
        if Y is None:
            Y = self.Y

        # Get the places of NaN values not supported by PyTorch models
        nans = list(set(np.where(np.isnan(X))[0]) | set(np.where(np.isnan(Y))[0]))

        # List of indices that present no nans
        indices = np.arange(len(X))
        not_nans_indices = np.delete(indices, nans)
        last = len(indices) - 1

        sequences = []

        if len(not_nans_indices) > 0:
            # Get the "jumps", i.e. where the the nan values appear
            jumps = np.concatenate([[True], np.diff(not_nans_indices) != 1, [True]])

            # Get the beginnings of all the sequences, correcting extreme values and adding 0 if needed
            beginnings = list(not_nans_indices[np.where(jumps[:-1])[0]])
            if 0 in beginnings:
                beginnings = beginnings[1:]
            if last in beginnings:
                beginnings = beginnings[:-1]
            if (0 in not_nans_indices) and (1 in not_nans_indices):
                beginnings = [0] + beginnings

            # Get the ends of all the sequences, correcting extreme values and adding the last value if needed
            ends = list(not_nans_indices[np.where(jumps[1:])[0]])
            if 0 in ends:
                ends = ends[1:]
            if last in ends:
                ends = ends[:-1]
            if (last in not_nans_indices) and (last - 1 in not_nans_indices):
                ends = ends + [last]

            # We should have the same number of series beginning and ending
            assert len(beginnings) == len(ends), "Something went wrong"

            # Bulk of the work: create starts and ends of sequences tuples
            for beginning, end in zip(beginnings, ends):
                # Add sequences from the start to the end, jumping with the wanted overlapping distance and ensuring
                # the required warm start length and minimum sequence length are respected
                sequences += [(beginning + self.overlapping_distance * x,
                               min(beginning + self.warm_start_length + self.maximum_sequence_length
                                   + self.overlapping_distance * x, end))
                    for x in range(math.ceil((end - beginning - self.warm_start_length
                                              - self.minimum_sequence_length) / self.overlapping_distance))]

        if inplace:
            self.sequences = sequences
        else:
            return sequences

    def get_sequences(self, X: pd.DataFrame = None, Y: pd.DataFrame = None) -> list:
        """
        Function to get tuple designing the beginning and end of sequences of data we can predict.
        This is needed because PyTorch models don't work with NaN values, so we need to only path
        sequences of data that don't contain any.

        If no sequences exist, it creates them.

        Args:
            X:          input data
            Y:          output data, i.e. labels

        Returns:
            All the sequences we can predict
        """

        # Create the corresponding name
        name = os.path.join(self.save_name, "sequences.pt")

        if X is None:
            X = self.X
        if Y is None:
            Y = self.Y

        if self.verbose > 0:
            print("\nTrying to load the predictable sequences, where the data has no missing values...")

        try:
            # Check the existence of the model
            assert os.path.exists(name), f"The file {name} doesn't exist."
            # Load the checkpoint
            checkpoint = torch.load(name, weights_only=False)
            # Put it into the model
            heating_sequences = checkpoint["heating_sequences"]
            cooling_sequences = checkpoint["cooling_sequences"]

            if self.verbose > 0:
                print("Found!")

        except AssertionError:
            if self.verbose > 0:
                print("Nothing found, building the sequences...")

            # Create the sequences
            if self.heating:
                X_ = X.copy()
                X_[np.where(X_[:, self.case_column] < 0.5)[0]] = np.nan
                heating_sequences = self._create_sequences(X=X_, Y=Y)
            else:
                heating_sequences = []

            if self.cooling:
                X_ = X.copy()
                X_[np.where(X_[:, self.case_column] > 0.5)[0]] = np.nan
                cooling_sequences = self._create_sequences(X=X_, Y=Y)
            else:
                cooling_sequences = []

            # Save the built list to be able to load it later and avoid the computation
            torch.save({"heating_sequences": heating_sequences, "cooling_sequences": cooling_sequences}, name)

        if self.verbose > 0:
            print(f"Number of sequences for the model {self.name}: {len(heating_sequences)} heating sequences and " f"{len(cooling_sequences)} cooling sequences.")

        # Return the sequences
        return heating_sequences, cooling_sequences

    def train_test_validation_separation(self, validation_percentage: float = 0.2, test_percentage: float = 0.0) -> None:
        """
        Function to separate the data into training and testing parts. The trick here is that
        the data is not actually split - this function actually defines the sequences of
        data points that are in the training/testing part.

        Args:
            validation_percentage:  Percentage of the data to keep out of the training process
                                    for validation
            test_percentage:        Percentage of data to keep out of the training process for
                                    the testing

        Returns:
            Nothing, in place definition of all the indices
        """

        # Sanity checks: the given inputs are given as percentage between 0 and 1
        if 1 <= validation_percentage <= 100:
            validation_percentage /= 100
            print("The train-test-validation separation rescaled the validation_percentage between 0 and 1")
        if 1 <= test_percentage <= 100:
            test_percentage /= 100
            print("The train-test-validation separation rescaled the test_percentage between 0 and 1")

        # Prepare the lists
        self.train_sequences = []
        self.validation_sequences = []
        self.test_sequences = []

        if self.verbose > 0:
            print("Creating training, validation and testing data")

        for sequences in [self.heating_sequences, self.cooling_sequences]:
            if len(sequences) > 0:
                # Given the total number of sequences, define aproximate separations between training
                # validation and testing sets
                train_validation_sep = int((1 - test_percentage - validation_percentage) * len(sequences))
                validation_test_sep = int((1 - test_percentage) * len(sequences))

                # Little trick to ensure training, validation and test sequences are completely distinct
                while True:
                    if (sequences[train_validation_sep - 1][1] < sequences[train_validation_sep][0]) | (train_validation_sep == 1):
                        break
                    train_validation_sep -= 1
                if test_percentage > 0.:
                    while True:
                        if (sequences[validation_test_sep - 1][1] < sequences[validation_test_sep][0]) | (validation_test_sep == 1):
                            break
                        validation_test_sep -= 1

                # Prepare the lists
                self.train_sequences += sequences[:train_validation_sep]
                self.validation_sequences += sequences[train_validation_sep:validation_test_sep]
                self.test_sequences += sequences[validation_test_sep:]

    def compute_zero_power(self):
        """
        Small helper function to compute the scaled value of zero power
        """

        # Scale the zero
        if self.dataset.is_normalized:
            min_ = self.dataset.min_.iloc[self.power_column]
            max_ = self.dataset.max_.iloc[self.power_column]
            zero = 0.8 * (0.0 - min_) / (max_ - min_) + 0.1

        elif self.dataset.is_standardized:
            mean = self.dataset.mean.iloc[self.power_column]
            std = self.dataset.std.iloc[self.power_column]
            zero = (0.0 - mean) / std

        else:
            zero = np.array([0.0] * len(self.rooms))

        return np.array(zero)

    def create_scalings(self):
        """
        Function to initialize good parameters for a, b, c and d, the key parameters of the structure.
        Intuition:
          - The room loses 1.5 degrees in 6h when the outside temperature is 25 degrees lower than
              the inside one (and some for losses to the neighboring room)
          - The room gains 2 degrees in 4h of heating

        Returns:
            The scaling parameters according to the data
        """

        parameter_scalings = {}

        if self.unit == 'DFAB':
            # DFAB power is in kW
            parameter_scalings['b'] = [1 / (2 / (self.dataset.max_ - self.dataset.min_).iloc[self.temperature_column]
                                           * 0.8 / 25 / 6 / 60 * self.dataset.interval).mean()]
            parameter_scalings['c'] = [1 / (2.5 / (self.dataset.max_ - self.dataset.min_).iloc[self.temperature_column]
                                           * 0.8 / 25 / 6 / 60 * self.dataset.interval).mean() / 10]

            parameter_scalings['a'] = [
                1 / (1 / (self.dataset.max_ - self.dataset.min_).iloc[self.temperature_column[i]] * 0.8  # Gain 2 degrees
                     / (self.dataset.max_['Thermal total power'] / (self.dataset.max_ - self.dataset.min_).iloc[
                            'Thermal total power'] * 0.8)  # With average power of 1500 W
                     / (10 * 60 / self.dataset.interval)) for i in range(len(self.rooms))]  # In 4h
            parameter_scalings['d'] = [
                1 / (1 / (self.dataset.max_ - self.dataset.min_).iloc[self.temperature_column[i]] * 0.8  # Gain 2 degrees
                     / (- self.dataset.min_['Thermal total power'] / (self.dataset.max_ - self.dataset.min_).iloc[
                            'Thermal total power'] * 0.8)  # With average power of 1000 W
                     / (10 * 60 / self.dataset.interval)) for i in range(len(self.rooms))]  # In 4h

        else:
            parameter_scalings['b'] = [1 / (1.5 / (self.dataset.max_ - self.dataset.min_).iloc[self.temperature_column]
                                           * 0.8 / 25 / 6 / 60 * self.dataset.interval).mean()]
            parameter_scalings['c'] = [1 / (1.5 / (self.dataset.max_ - self.dataset.min_).iloc[self.temperature_column]
                                           * 0.8 / 25 / 6 / 60 * self.dataset.interval).mean() / 10]

            # needed condition to make sure to deal with data in Watts and kWs
            if (self.dataset.max_ - self.dataset.min_).iloc[self.power_column[0]] > 100:
                parameter_scalings['a'] = [
                    1 / (2 / (self.dataset.max_ - self.dataset.min_).iloc[self.temperature_column[i]] * 0.8  # Gain 2 degrees
                         / (1000 / (self.dataset.max_ - self.dataset.min_).iloc[
                                self.power_column[i]] * 0.8)  # With average power of 1.5 kW
                         / (4 * 60 / self.dataset.interval)) for i in range(len(self.rooms))] # in 4h
                parameter_scalings['d'] = [
                    1 / (2 / (self.dataset.max_ - self.dataset.min_).iloc[self.temperature_column[i]] * 0.8  # Gain 2 degrees
                         / (1000 / (self.dataset.max_ - self.dataset.min_).iloc[
                                self.power_column[i]] * 0.8)  # With average power of 1.5 kW
                         / (4 * 60 / self.dataset.interval)) for i in range(len(self.rooms))] # in 4h
            else:
                parameter_scalings['a'] = [
                    1 / (2 / (self.dataset.max_ - self.dataset.min_).iloc[self.temperature_column[i]] * 0.8  # Gain 2 degrees
                         / (self.dataset.max_.iloc[self.power_column[i]] * 3/2 / (self.dataset.max_ - self.dataset.min_).iloc[
                                self.power_column[i]] * 0.8)  # With average power of 1.5 kW
                         / (4 * 60 / self.dataset.interval)) for i in range(len(self.rooms))]  # in 4h
                parameter_scalings['d'] = [
                    1 / (2 / (self.dataset.max_ - self.dataset.min_).iloc[self.temperature_column[i]] * 0.8  # Gain 2 degrees
                         / (self.dataset.max_.iloc[self.power_column[i]] * 3/2 / (self.dataset.max_ - self.dataset.min_).iloc[
                                self.power_column[i]] * 0.8)  # With average power of 1.5 kW
                         / (4 * 60 / self.dataset.interval)) for i in range(len(self.rooms))]  # in 4h

        return parameter_scalings

    def get_normalization_variables(self):
        """
        Function to get the minimum and the amplitude of some variables in the data. In particular, we need
        that for the room temperature, the outside temperature and the neighboring room temperature.
        This is used by the physics-inspired network to unnormalize the predictions.
        """
        normalization_variables = {}
        normalization_variables['Room'] = [torch.Tensor(self.dataset.min_.iloc[self.temperature_column].values).to(self.device),
                                           torch.Tensor((self.dataset.max_ - self.dataset.min_).iloc[self.temperature_column].values).to(self.device)]
        if self.neigh_column is not None:
            normalization_variables['Neigh'] = [torch.Tensor(self.dataset.min_.iloc[self.neigh_column].values).to(self.device),
                                           torch.Tensor((self.dataset.max_ - self.dataset.min_).iloc[self.neigh_column].values).to(self.device)]
        normalization_variables['Out'] = [torch.Tensor([self.dataset.min_.iloc[self.out_column]]).to(self.device),
                                          torch.Tensor([(self.dataset.max_ - self.dataset.min_).iloc[self.out_column]]).to(self.device)]
        return normalization_variables

    def batch_iterator(self, iterator_type: str = "train", batch_size: int = None, shuffle: bool = True) -> None:
        """
        Function to create batches of the data with the wanted size, either for training,
        validation, or testing

        Args:
            iterator_type:  To know if this should handle training, validation or testing data
            batch_size:     Size of the batches
            shuffle:        Flag to shuffle the data before the batches creation

        Returns:
            Nothing, yields the batches to be then used in an iterator
        """

        # Firstly control that the training sequences exist - create them otherwise
        if self.train_sequences is None:
            self.train_test_validation_separation()
            print("The Data was not separated in train, validation and test --> the default 70%-20%-10% was used")

        # If no batch size is given, define it as the default one
        if batch_size is None:
            batch_size = self.batch_size

        # Copy the indices of the correct type (without first letter in case of caps)
        if "rain" in iterator_type:
            sequences = self.train_sequences
        elif "alidation" in iterator_type:
            sequences = self.validation_sequences
        elif "est" in iterator_type:
            sequences = self.test_sequences
        else:
            raise ValueError(f"Unknown type of batch creation {iterator_type}")

        # Shuffle them if wanted
        if shuffle:
            np.random.shuffle(sequences)

        # Define the right number of batches according to the wanted batch_size - taking care of the
        # special case where the indicies ae exactly divisible by the batch size, which can induce
        # an additional empty batch breaking the simulation down the line
        n_batches = int(np.ceil(len(sequences) / batch_size))

        # Iterate to yield the right batches with the wanted size
        for batch in range(n_batches):
            yield sequences[batch * batch_size: (batch + 1) * batch_size]

    def build_input_output_from_sequences(self, sequences: list):
        """
        Input and output generator from given sequences of indices corresponding to a batch.

        Args:
            sequences: sequences of the batch to prepare

        Returns:
            batch_x:    Batch input of the model
            batch_y:    Targets of the model, the temperature and the power
        """

        # Ensure the given sequences are a list of list, not only one list
        if type(sequences) == tuple:
            sequences = [sequences]

        # Iterate over the sequences to build the input in the right form
        input_tensor_list = [torch.Tensor(self.X[sequence[0]: sequence[1], :].copy()) for sequence in sequences]

        # Prepare the output for the temperature and power consumption
        output_tensor_list = [torch.Tensor(self.Y[sequence[0]: sequence[1], :].copy()) for sequence in
                                    sequences]


        # Build the final results by taking care of the batch_size=1 case
        if len(sequences) > 1:
            batch_x = pad_sequence(input_tensor_list, batch_first=True, padding_value=0)
            batch_y = pad_sequence(output_tensor_list, batch_first=True, padding_value=0)
        else:
            batch_x = input_tensor_list[0].view(1, input_tensor_list[0].shape[0], -1)
            batch_y = output_tensor_list[0].view(1, output_tensor_list[0].shape[0], -1)

        # Return everything
        return batch_x.to(self.device), batch_y.to(self.device)

    def predict(self, sequences: Union[list, int] = None, data: torch.Tensor = None, mpc_mode: bool = False):
        """
        Function to predict batches of "sequences", i.e. it creates batches of input and output of the
        given sequences you want to predict and forwards them through the network

        Args:
            sequences:  Sequences of the data to predict
            data:       Alternatively, data to predict, a tuple of tensors with the X and Y (if there is no
                          Y just put a vector of zeros with the right output size)
            mpc_mode:   Flag to pass to the MPC mode and return D and E separately

        Returns:
            The predictions and the true output
        """

        return_y = True
        if sequences is not None:
            # Ensure the given sequences are a list of list, not only one list
            if type(sequences) == tuple:
                sequences = [sequences]

            # Build the input and output
            batch_x, batch_y = self.build_input_output_from_sequences(sequences=sequences)

        elif data is not None:
            if isinstance(data, tuple):
                if len(data[0].shape) == 3:
                    batch_x = data[0].reshape(data[0].shape[0], data[0].shape[1], -1)
                    batch_y = data[1].reshape(data[0].shape[0], data[0].shape[1], len(self.rooms))
                else:
                    batch_x = data[0].reshape(1, data[0].shape[0], -1)
                    batch_y = data[1].reshape(1, data[0].shape[0], len(self.rooms))
            else:
                if len(data.shape) == 3:
                    batch_x = data.reshape(data.shape[0], data.shape[1], -1)
                else:
                    batch_x = data.reshape(1, data.shape[0], -1)
                return_y = False

        else:
            raise ValueError("Either sequences or data must be provided to the `predict` function")

        predictions = torch.zeros((batch_x.shape[0], batch_x.shape[1], len(self.rooms))).to(self.device)
        states = None

        # Iterate through the sequences of data to predict each step, replacing the true power and temperature
        # values with the predicted ones each time
        for i in range(batch_x.shape[1]):
            # Predict the next output and store it
            pred, states = self.model(batch_x[:, i, :], states, warm_start=i<self.warm_start_length)
            predictions[:, i, :] = pred

        if return_y:
            return predictions, batch_y
        else:
            return predictions

    def scale_back_predictions(self, sequences: Union[list, int] = None, data: torch.Tensor = None):
        """
        Function preparing the data for analyses: it predicts the wanted sequences and returns the scaled
        predictions and true_data

        Args:
            sequences:  Sequences to predict
            data:       Alternatively, data to predict, a tuple of tensors with the X and Y (if there is no
                          Y just put a vector of zeros with the right output size)

        Returns:
            The predictions and the true data
        """

        # Compute the predictions and get the true data out of the GPU
        predictions, true_data = self.predict(sequences=sequences, data=data)
        predictions = predictions.cpu().detach().numpy()
        true_data = true_data.cpu().detach().numpy()

        # Reshape the data for consistency with the next part of the code if only one sequence is given
        if sequences is not None:
            # Reshape when only 1 sequence given
            if type(sequences) == tuple:
                sequences = [sequences]

        elif data is not None:
            sequences = [0]

        else:
            raise ValueError("Either sequences or data must be provided to the `scale_back_predictions` function")

        if len(predictions.shape) == 2:
            predictions = predictions.reshape(1, predictions.shape[0], -1)
            true_data = true_data.reshape(1, true_data.shape[0], -1)

        # Scale the data back
        cols = self.dataset.Y_columns[:-1] if self.predict_power else self.dataset.Y_columns[:-2]
        truth = true_data.reshape(true_data.shape[0], true_data.shape[1], -1)
        true = np.zeros_like(predictions)

        if self.dataset.is_normalized:
            for i, sequence in enumerate(sequences):
                predictions[i, :, :] = inverse_normalize(data=predictions[i, :, :],
                                                         min_=self.dataset.min_[self.dataset.Y_columns],
                                                         max_=self.dataset.max_[self.dataset.Y_columns])
                true[i, :, :] = inverse_normalize(data=truth[i, :, :],
                                                       min_=self.dataset.min_[self.dataset.Y_columns],
                                                       max_=self.dataset.max_[self.dataset.Y_columns])
        elif self.dataset.is_standardized:
            for i, sequence in enumerate(sequences):
                predictions[i, :, :] = inverse_standardize(data=predictions[i, :, :],
                                                           mean=self.dataset.mean[self.dataset.Y_columns],
                                                           std=self.dataset.std[self.dataset.Y_columns])
                true[i, :, :] = inverse_standardize(data=truth[i, :, :],
                                                         mean=self.dataset.mean[self.dataset.Y_columns],
                                                         std=self.dataset.std[self.dataset.Y_columns])

        return predictions, true

    def fit(self, n_epochs: int = None, print_each: int = 5) -> None:
        """
        General function fitting a model for several epochs, training and evaluating it on the data.

        Args:
            n_epochs:         Number of epochs to fit the model, if None this takes the default number
                                defined by the parameters
            n_batches_print:  Control how many batches to print per epoch

        Returns:
            Nothing
        """

        self.times.append(time.time())

        if self.verbose > 0:
            print("\nTraining starts!")

        # If no special number of epochs is given, take the default one
        if n_epochs is None:
            n_epochs = self.n_epochs

        # Define the best loss, taking the best existing one or a very high loss
        best_loss = np.min(self.validation_losses) if len(self.validation_losses) > 0 else np.inf

        # Assess the number of epochs the model was already trained on to get nice prints
        trained_epochs = len(self.train_losses)

        for epoch in range(trained_epochs, trained_epochs + n_epochs):

            if self.verbose > 0:
                print(f"\nTraining epoch {epoch + 1}...")

            # Start the training, define a list to retain the training losses along the way
            self.model.train()
            train_losses = []
            train_sizes = []

            # Adjust the learning rate if wanted
            if self.decrease_learning_rate:
                self.adjust_learning_rate(epoch=epoch)

            # Create training batches and run through them, using the batch_iterator function, which has to be defined
            # independently for each subclass, as different types of data are handled differently
            for num_batch, batch_sequences in enumerate(self.batch_iterator(iterator_type="train")):

                # Compute the loss of the batch and store it
                loss = self.compute_loss(batch_sequences)

                # Compute the gradients and take one step using the optimizer
                loss.backward()
                #for p in self.model.named_parameters():
                #    if (p[1].grad is not None):
                #        print(p[0], ":", p[1].grad.norm())
                self.optimizer.step()
                self.optimizer.zero_grad()
                train_losses.append(float(loss))
                train_sizes.append(len(batch_sequences))

                # Regularly print the current state of things
                if (self.verbose > 1) & (num_batch % print_each == print_each - 1):
                    print(f"Loss batch {num_batch + 1}: {float(loss):.2E}")

            # Compute the average loss of the training epoch and print it
            train_loss = sum([l*s for l,s in zip(train_losses, train_sizes)]) / sum(train_sizes)
            print(f"Average training loss after {epoch + 1} epochs: {train_loss:.2E}")
            self.train_losses.append(train_loss)

            # Start the validation, again defining a list to recall the losses
            if self.verbose > 0:
                print(f"Validation epoch {epoch + 1}...")
            validation_losses = []
            _validation_losses = []
            validation_sizes = []
            _validation_sizes = []

            # Create validation batches and run through them. Note that we use larger batches
            # to accelerate it a bit, and there is no need to shuffle the indices
            for num_batch, batch_sequences in enumerate(self.batch_iterator(iterator_type="validation", batch_size=2 * self.batch_size, shuffle=False)):

                # Compute the loss, in the torch.no_grad setting: we don't need the model to
                # compute and use gradients here, we are not training
                if 'PiNN' not in self.name:
                    self.model.eval()
                    with torch.no_grad():
                        loss = self.compute_loss(batch_sequences)
                        validation_losses.append(float(loss))
                        validation_sizes.append(len(batch_sequences))
                        # Regularly print the current state of things
                        if (self.verbose > 1) & (num_batch % (print_each//2) == (print_each//2) - 1):
                            print(f"Loss batch {num_batch + 1}: {float(loss):.2E}")

                else:
                    self.model.train()
                    loss = self.compute_loss(batch_sequences)
                    validation_losses.append(float(loss))
                    validation_sizes.append(len(batch_sequences))
                    # Regularly print the current state of things
                    if (self.verbose > 1) & (num_batch % (print_each//2) == (print_each//2) - 1):
                        print(f"Loss batch {num_batch + 1}: {float(loss):.2E}")
                    self.model.eval()
                    with torch.no_grad():
                        loss = self._compute_loss(batch_sequences)
                        _validation_losses.append(float(loss))
                        _validation_sizes.append(len(batch_sequences))
                        # Regularly print the current state of things
                        if (self.verbose > 1) & (num_batch % (print_each//2) == (print_each//2) - 1):
                            print(f"Loss batch {num_batch + 1}: {float(loss):.2E}")

            # Compute the average validation loss of the epoch and print it
            validation_loss = sum([l*s for l,s in zip(validation_losses, train_sizes)]) / sum(validation_sizes)
            self.validation_losses.append(validation_loss)
            print(f"Average validation loss after {epoch + 1} epochs: {validation_loss:.2E}")

            if 'PiNN' in self.name:
                _validation_loss = sum([l*s for l,s in zip(_validation_losses, train_sizes)]) / sum(_validation_sizes)
                self._validation_losses.append(_validation_loss)
                if self.verbose > 0:
                    print(f"Average accuracy validation loss after {epoch + 1} epochs: {_validation_loss:.2E}")

            # Timing information
            self.times.append(time.time())
            if self.verbose > 0:
                print(f"Time elapsed for the epoch: {format_elapsed_time(self.times[-2], self.times[-1])}"
                      f" - for a total training time of {format_elapsed_time(self.times[0], self.times[-1])}")

            # Save parameters
            if 'PCNN' in self.module:
                p = self.model.E_parameters
                self.a.append(p[0])
                self.b.append(p[1])
                self.c.append(p[2])
                self.d.append(p[3])

            # Save last and possibly best model
            self.save(name_to_add="last", verbose=0)

            if validation_loss < best_loss:
                self.save(name_to_add="best", verbose=1)
                best_loss = validation_loss

        if self.verbose > 0:
            best_epoch = np.argmin([x for x in self.validation_losses])
            print(f"\nThe best model was obtained at epoch {best_epoch + 1} after training for " f"{trained_epochs + n_epochs} epochs")

    def adjust_learning_rate(self, epoch: int) -> None:
        """
        Custom function to decrease the learning rate along the training

        Args:
            epoch:  Epoch of the training

        Returns:
            Nothing, modifies the optimizer in place
        """

        for param_group in self.optimizer.param_groups:
            param_group["lr"] = param_group["lr"] * 0.997

    def compute_loss(self, sequences: list):
        """
        Custom function to compute the loss of a batch of sequences.

        Args:
            sequences: The sequences in the batch

        Returns:
            The loss
        """

        predictions, batch_y = self.predict(sequences=sequences)
        return self.loss(predictions, batch_y)

    def save(self, name_to_add: str = None, verbose: int = 0):
        """
        Function to save a PyTorch model: Save the state of all parameters, as well as the one of the
        optimizer. We also recall the losses for ease of analysis.

        Args:
            name_to_add:    Something to save a unique model

        Returns
            Nothing, everything is done in place and stored in the parameters
        """

        if verbose > 0:
            print(f"\nSaving the new {name_to_add} model!")

        if name_to_add is not None:
            save_name = os.path.join(self.save_name, f"{name_to_add}_model.pt")
        else:
            save_name = os.path.join(self.save_name, "model.pt")

        torch.save(
            {
                "model_state_dict": self.model.state_dict(),
                "optimizer_state_dict": self.optimizer.state_dict(),
                "train_sequences": self.train_sequences,
                "validation_sequences": self.validation_sequences,
                "test_sequences": self.test_sequences,
                "train_losses": self.train_losses,
                "validation_losses": self.validation_losses,
                "_validation_losses": self._validation_losses,
                "test_losses": self.test_losses,
                "times": self.times,
                "a": self.a,
                "b": self.b,
                "c": self.c,
                "d": self.d,
                "warm_start_length": self.warm_start_length,
                "maximum_sequence_length": self.maximum_sequence_length,
                "feed_input_through_nn": self.feed_input_through_nn,
                "input_nn_hidden_sizes": self.input_nn_hidden_sizes,
                "lstm_hidden_size": self.lstm_hidden_size,
                "lstm_num_layers": self.lstm_num_layers,
                "output_nn_hidden_sizes": self.output_nn_hidden_sizes,
            },
            save_name,
        )

    def load(self, load_last: bool = False):
        """
        Function trying to load an existing model, by default the best one if it exists. But for training purposes,
        it is possible to load the last state of the model instead.

        Args:
            load_last:  Flag to set to True if the last model checkpoint is wanted, instead of the best one

        Returns:
             Nothing, everything is done in place and stored in the parameters.
        """

        if load_last:
            save_name = os.path.join(self.save_name, "last_model.pt")
        else:
            save_name = os.path.join(self.save_name, "best_model.pt")

        if self.verbose > 0:
            print("\nTrying to load a trained model...")
        try:
            # Build the full path to the model and check its existence

            assert os.path.exists(save_name), f"The file {save_name} doesn't exist."

            # Load the checkpoint
            checkpoint = torch.load(save_name, map_location=lambda storage, loc: storage, weights_only=False)

            # Put it into the model
            self.model.load_state_dict(checkpoint["model_state_dict"])
            self.optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
            if torch.cuda.is_available():
                for state in self.optimizer.state.values():
                    for k, v in state.items():
                        if isinstance(v, torch.Tensor):
                            state[k] = v.cuda()
            self.train_sequences = checkpoint["train_sequences"]
            self.validation_sequences = checkpoint["validation_sequences"]
            self.test_sequences = checkpoint["test_sequences"]
            self.train_losses = checkpoint["train_losses"]
            self.validation_losses = checkpoint["validation_losses"]
            self.test_losses = checkpoint["test_losses"]
            self.times = checkpoint["times"]
            self.a = checkpoint["a"]
            self.b = checkpoint["b"]
            self.c = checkpoint["c"]
            self.d = checkpoint["d"]
            self.warm_start_length = checkpoint["warm_start_length"]
            self.maximum_sequence_length = checkpoint["maximum_sequence_length"]
            self.feed_input_through_nn = checkpoint["feed_input_through_nn"]
            self.input_nn_hidden_sizes = checkpoint["input_nn_hidden_sizes"]
            self.lstm_hidden_size = checkpoint["lstm_hidden_size"]
            self.lstm_num_layers = checkpoint["lstm_num_layers"]
            self.output_nn_hidden_sizes = checkpoint["output_nn_hidden_sizes"]

            # Print the current status of the found model
            if self.verbose > 0:
                print(f"Found!\nThe model has been fitted for {len(self.train_losses)} epochs already, "
                      f"with loss {np.min(self.validation_losses): .5f}.")
                print(f"It contains {len(self.train_sequences)} training sequences and "
                      f"{len(self.validation_sequences)} validation sequences.\n")

        # Otherwise, keep an empty model, return nothing
        except AssertionError:
            print("\nNo existing model was found!\n")
